(** ** Products in presheafs *)

Section PresheafProducts.
  Context {A : cat}.
  Implicit Types (F G : PRESHEAF A).

  Program Definition PProd F G : PRESHEAF A :=
    mk_functor (fun a => (F a * G a)%type) (fun a b f p => (map F f p.1, map G f p.2)) _ _.
  Next Obligation. fext=> -[x y] /=. by csimpl. Qed.
  Next Obligation. fext=> -[x y] /=. csimpl=>/=. by csimpl. Qed.

  Program Definition nat_pi1 {F G} : hom (PProd F G) F :=
    mk_nat (fun a p => p.1) _.

  Program Definition nat_pi2 {F G} : hom (PProd F G) G :=
    mk_nat (fun a p => p.2) _.

  Program Definition nat_pair {F G H} (eps : hom F G) (eta : hom F H) : hom F (PProd G H) :=
    mk_nat (fun a x => (eps a x, eta a x)) _.
  Next Obligation. fext=> c. csimpl=> /=. by csimpl. Qed.

  Lemma nat_pair1 F G H (eps : hom F G) (eta : hom F H) :
    nat_pair eps eta ∘ nat_pi1 = eps.
  Proof. exact: nat_eq. Qed.

  Lemma nat_pair2 F G H (eps : hom F G) (eta : hom F H) :
    nat_pair eps eta ∘ nat_pi2 = eta.
  Proof. exact: nat_eq. Qed.

  Lemma nat_pair_eta F G H (eta : nat F (PProd G H)) :
    nat_pair (eta ∘ nat_pi1) (eta ∘ nat_pi2) = eta.
  Proof.
    apply: nat_eq. fext=> a x /=. csimpl. by destruct (eta a x).
  Qed.
End PresheafProducts.

Instance presheaf_has_products (A : cat) : has_products (PRESHEAF A) :=
  mk_has_products nat_pair1 nat_pair2 nat_pair_eta.

(** ** Exponentials in presheafs

  hom (a × b) c <~> hom a c^b
  (natural in a and c)

hlam : hom (a × b) c -> hom a c^b
happly : hom (c^b × b) c

hlam^-1 : hom a c^b -> hom (a × b) c
hlam^-1 f = ⟨ π₁ ∘ f, π₂ ⟩ ∘ eval

hlam (hlam^-1 f) = hlam (⟨ π₁ ∘ f, π₂ ⟩ ∘ eval) =! f
hlam^-1 (hlam f) = ⟨ π₁ ∘ hlam f, π₂ ⟩ ∘ eval =! f

naturality of hlam in a:
forall a a' (f : hom a a')

hlam (⟨ π₁ ∘ f, π₂ ⟩ ∘ g) = f ∘ hlam g

so:
  hlam (hlam^-1 f) = hlam (⟨ π₁ ∘ f, π₂ ⟩ ∘ eval)
                   = f ∘ hlam happly
  suffices to have "hlam happly = id" (the eta law)

functoriality of c^b in c:

map : hom c c' -> hom c^b c'^b
map f := hlam (eval ∘ f)

map id = hlam (eval ∘ id) = hlam eval = id
  map f ∘ map g
= hlam (eval ∘ f) ∘ hlam (eval ∘ g)
= hlam (⟨ π₁ ∘ hlam (eval ∘ f), π₂ ⟩ ∘ eval ∘ g)
= hlam ( eval ∘ f ∘ g )
= map (f ∘ g)

contravariant functoriality of c^b in b

map : hom b b' -> hom c^b' c^b
map f = hlam (⟨ π₁, f ∘ π₂ ⟩ ∘ eval)

  map id
= hlam (⟨ π₁, id ∘ π₂ ⟩ ∘ eval)
= hlam eval
= id

  map f ∘ map g
= hlam (⟨ π₁, f ∘ π₂ ⟩ ∘ eval) ∘ hlam (⟨ π₁, g ∘ π₂ ⟩ ∘ eval)
= hlam (⟨ π₁ ∘ hlam (⟨ π₁, f ∘ π₂ ⟩ ∘ eval), π₂ ⟩ ∘ ⟨ π₁, g ∘ π₂ ⟩ ∘ eval)
= hlam (⟨ π₁, g ∘ π₂ ⟩ ∘ ⟨ π₁ ∘ hlam (⟨ π₁, f ∘ π₂ ⟩ ∘ eval), π₂ ⟩ ∘ eval)
= hlam (⟨ π₁, g ∘ π₂ ⟩ ∘ ⟨ π₁, f ∘ π₂ ⟩ ∘ eval)
= hlam (⟨ π₁, g ∘ f ∘ π₂ ⟩)
= map (g ∘ f)

naturality of hlam in c

f : hom (a × b) c
g : hom c c'

hlam (f ∘ g) =! hlam f ∘ hlam (eval ∘ g)

= hlam (⟨ π₁ ∘ hlam (eval ∘ f), π₂ ⟩ ∘ eval ∘ g)
= hlam (eval ∘ f ∘ g)

naturality of hlam in b

f : hom (a × b) c
g : hom b' b

hlam (⟨ π₁, g ∘ π₂ ⟩ ∘ f) =! hlam f ∘ map g
 *)

Section PresheafExponentials.
  Variable A : cat.
  Implicit Types (F G H : PRESHEAF A).

  Local Obligation Tactic := idtac.

  Program Definition PExp F G : PRESHEAF A :=
    mk_functor (fun (a : A) => hom (Prod (y a) G) F) (fun a b (f : A b a) g => hfirst (ymap f) ∘ g) _ _.
  Next Obligation.
    move=> F G a. fext=> f /=. csimpl. rewrite-> ymap1. csimpl.
    move: (@hpair_eta_id (PRESHEAF A) (presheaf_has_products A)).
    cbn. move=> E. rewrite-> E. exact: nat_idL.
  Qed.
  Next Obligation.
    move=> F G a b c f g. fext=> h. csimpl.
      by rewrite <-ymapC, <-hfirst_comp, <- compA.
  Qed.

  Program Definition nat_happly F G : hom (Prod (PExp F G) G) F :=
    mk_nat (fun a etap => etap.1 a (id, etap.2)) _.
  Next Obligation.
    move=> F G a b f. fext=> -[eta x] /=. csimpl=>/=. by csimpl=>/=.
  Qed.

  Program Definition nat_hlam F G H (f : hom (Prod F G) H) : hom F (PExp H G) :=
    mk_nat (fun a x => mk_nat (fun b p => f b (map F p.1 x, p.2)) _) _.
  Next Obligation.
    move=> F G H f a x b c g. fext=> -[/=h y]. csimpl=> /=. by csimpl.
  Qed.
  Next Obligation.
    move=> F G H f a b g. fext=>/=x. csimpl=>/=. apply: nat_eq.
    cbn. fext=> c -[/=h y]. by csimpl=>/=.
  Qed.

  Lemma nat_hlam_beta F G H (f : hom (Prod F G) H) :
    hfirst (nat_hlam f) ∘ nat_happly H G = f.
  Proof.
    apply: nat_eq; fext=> a -[x y] /=. csimpl=>/=. by csimpl.
  Qed.

  Lemma nat_hlam_eta F G :
    nat_hlam (nat_happly F G) = id.
  Proof.
    apply: nat_eq; fext=> a /= eta. apply: nat_eq; fext=> b [/=f y]. by csimpl=>/=.
  Qed.

  Lemma nat_hlam_nat F G H I (f : hom F G) (g : hom (Prod G H) I) :
    nat_hlam (hfirst f ∘ g) = f ∘ nat_hlam g.
  Proof.
    apply: nat_eq; fext=> a x. apply: nat_eq; fext=> b [/=h y].
      by csimpl=> /=.
  Qed.
End PresheafExponentials.

Instance presheaf_has_exponentials A : @has_exponentials (PRESHEAF A) _ :=
  mk_has_exponentials (@nat_hlam_beta A) (@nat_hlam_eta A) (@nat_hlam_nat A).