(** * Categories

  Categories key on both their type of objects as well as the type of homomorphisms.
  This enables us to use categorical compositions (f ∘ g) in more contexts, but note
  that it also mean that we have to use aliases to obtain canonical structures on
  many types.
 *)
Require Import base.
Set Universe Polymorphism.

Module Cat.
  Polymorphic Cumulative Record mixin_of (T : Type) (hom : T -> T -> Type) := Mixin {
    id : forall a, hom a a;
    comp : forall a b c, hom a b -> hom b c -> hom a c;
    _ : forall a b (f : hom a b), comp (id a) f = f;
    _ : forall a b (f : hom a b), comp f (id b) = f;
    _ : forall a b c d (f : hom a b) (g : hom b c) (h : hom c d),
        comp (comp f g) h = comp f (comp g h)
  }.
  Notation class_of := mixin_of.

  Section ClassDef.
    Polymorphic Cumulative Structure type := Pack {sort; hom; _ : @class_of sort hom; sort' : Type; _ : sort' -> sort' -> Type }.
    Local Coercion sort : type >-> Sortclass.
    Local Coercion hom  : type >-> Funclass.

    Variable (T : Type) (hom : T -> T -> Type) (cT : type).
    Definition class := let: Pack _ _ c _ _ := cT return @class_of cT cT in c.

    Definition pack c := @Pack T hom c T hom.
    Definition clone c of phant_id class c := @Pack T hom c T hom.
  End ClassDef.
  
  Module Exports.
    Coercion sort : type >-> Sortclass.
    Coercion hom  : type >-> Funclass.

    Notation cat := type.
    Notation CatMixin := Mixin.
    Notation Cat T hom m := (@pack T hom m).

    Notation "[ 'catMixin' 'of' hom ]" := (class _ : @mixin_of _ hom)
      (at level 0, format "[ 'catMixin'  'of'  hom ]") : form_scope.
    Notation "[ 'cat' 'of' hom 'for' C ]" := (@clone _ hom C _ idfun)
      (at level 0, format "[ 'cat'  'of'  hom  'for'  C ]") : form_scope.
    Notation "[ 'cat' 'of' hom ]" := (@clone _ hom _ _ (fun x => x))
      (at level 0, format "[ 'cat'  'of'  hom ]") : form_scope.
  End Exports.
End Cat.
Export Cat.Exports.

(** ** Operations and Laws *)

Definition hom {T : cat} : T -> T -> Type := T.
Arguments hom {T} x y.

Definition cid {T : cat} {a : T} : hom a a :=
  Cat.id (Cat.class T) a.

Definition comp {T : cat} : forall a b c, hom a b -> hom b c -> hom a c :=
  @Cat.comp _ _ (Cat.class T).
Arguments comp {T a b c} f g.
Notation "f ∘ g" := (comp f g) : cat_scope.
Open Scope cat_scope.

Section Lemmas.
  Context {C : cat}.
  Implicit Types (a b c d : C).

  Lemma idL a b (f : hom a b) : cid ∘ f = f.
  Proof. by case: C a b f => /=T hom[]. Qed.

  Lemma idR a b (f : hom a b) : f ∘ cid = f.
  Proof. by case: C a b f => /=T hom[]. Qed.

  Lemma compA a b c d (f : hom a b) (g : hom b c) (h : hom c d) :
    (f ∘ g) ∘ h = f ∘ (g ∘ h).
  Proof. by case: C a b c d f g h => /=T hom[]. Qed.
End Lemmas.

(** ** Category of Sets *)

Section SetCategory.
  Definition arr A B := A -> B.

  Definition arr_id A : arr A A := idfun.
  Definition arr_comp A B C (f : arr A B) (g : arr B C) := f \; g.

  Canonical set_catMixin := @CatMixin Type arr arr_id arr_comp
    (fun _ _ _ => erefl) (fun _ _ _ => erefl) (fun _ _ _ _ _ _ _ => erefl).
  Canonical SET := Eval hnf in Cat Type arr set_catMixin.

  Lemma set_compE A B C (f : A -> B) (g : B -> C) (x : A) :
    ((f : arr A B) ∘ g) x = g (f x).
  Proof. by []. Qed.

  Lemma set_idE A x : cid (T := SET) (a := A) x = x.
  Proof. by []. Qed.
End SetCategory.

(** ** Opposite Category *)

Section OpCategory.
  Variable A : cat.
  Definition op : Type := A.
  Definition op_hom (a b : op) : Type := hom b a.

  Definition op_id (a : op) : op_hom a a := cid.
  Definition op_comp (a b c : op) (f : op_hom a b) (g : op_hom b c) := g ∘ f.

  Lemma op_idL a b (f : op_hom a b) : op_comp (op_id a) f = f.
  Proof. exact: idR. Qed.

  Lemma op_idR a b (f : op_hom a b) : op_comp f (op_id b) = f.
  Proof. exact: idL. Qed.

  Lemma op_compA a b c d (f : op_hom a b) (g : op_hom b c) (h : op_hom c d) :
    op_comp (op_comp f g) h = op_comp f (op_comp g h).
  Proof. unfold op_comp. by rewrite-> compA. Qed.

  Canonical op_catMixin := CatMixin op_idL op_idR op_compA.
  Canonical OP := Eval hnf in Cat op op_hom op_catMixin.
End OpCategory.

Notation "C '^op'" := (op C) (at level 2) : cat_scope.

(** ** Automation *)

(* unfold all occurrences of op in the goal *)
Ltac op_simpl :=
  match goal with
  | [|- context[@Cat.sort (OP ?A)]] =>
    change (@Cat.sort (OP A)) with (@Cat.sort A)
  | [|- context[@Cat.hom (OP ?A) ?a ?b]] =>
    change (@Cat.hom (OP A) a b) with (@Cat.hom A b a)
  | [|- context[@cid (OP ?A)]] =>
    change (@cid (OP A)) with (@cid A)
  | [|- context[@comp (OP ?A) ?a ?b ?c ?f ?g]] =>
    change (@comp (OP A) a b c f g) with (@comp A c b a g f)
  end.

Hint Rewrite @idL @idR @compA : cat.
Hint Rewrite set_compE set_idE : cat.
(*Ltac csimpl := try rewrite_strat (topdown (hints cat)).*)
Ltac csimpl := repeat first [op_simpl|autorewrite with cat|move=>/=].
