(** * Locally Ordered Categories

  A 2-poset (or locally partially ordered category, or skeletal 2-category, or (1,2)-category)
  is a category with a partial order on each of its homsets.
 *)
Require Import base ord.

Set Universe Polymorphism.

Module Cat.
  Polymorphic Cumulative Record mixin_of (T : Type) := Mixin {
    hom : T -> T -> ordType;
    id : forall a, hom a a;
    comp : forall a b c, hom a b -> hom b c -> hom a c;
    _ : forall a b (f : hom a b), comp (id a) f = f;
    _ : forall a b (f : hom a b), comp f (id b) = f;
    _ : forall a b c d (f : hom a b) (g : hom b c) (h : hom c d),
        comp (comp f g) h = comp f (comp g h);
    _ : forall a b c, Proper (ord_op ++> ord_op ++> ord_op) (@comp a b c)
  }.
  Notation class_of := mixin_of.

  Section ClassDef.
    Polymorphic Cumulative Structure type := Pack {sort; _ : class_of sort; _ : Type }.
    Local Coercion sort : type >-> Sortclass.

    Variable (T : Type) (cT : type).
    Definition class := let: Pack _ c _ := cT return class_of cT in c.

    Definition pack c := @Pack T c T.
    Definition clone c of phant_id class c := @Pack T c T.
  End ClassDef.
  
  Module Exports.
    Coercion sort : type >-> Sortclass.
    Notation pcat := type.
    Notation PCatMixin := Mixin.
    Notation PCat T m := (@pack T m).

    Notation "[ 'pcatMixin' 'of' T ]" := (class _ : mixin_of T)
      (at level 0, format "[ 'pcatMixin'  'of'  T ]") : form_scope.
    Notation "[ 'pcat' 'of' T 'for' C ]" := (@clone T C _ idfun)
      (at level 0, format "[ 'pcat'  'of'  T  'for'  C ]") : form_scope.
    Notation "[ 'pcat' 'of' T ]" := (@clone T _ _ id)
      (at level 0, format "[ 'pcat'  'of'  T ]") : form_scope.
  End Exports.
End Cat.
Export Cat.Exports.

Definition hom {T : pcat} : T -> T -> Type := Cat.hom (Cat.class T).
Arguments hom {T} x y.

Definition id {T : pcat} : forall a, hom a a :=
  Cat.id (Cat.class T).
Arguments id {T a}.

Definition comp {T : pcat} : forall a b c, hom a b -> hom b c -> hom a c :=
  @Cat.comp _ (Cat.class T).
Arguments comp {T a b c} f g.
Notation "f ∘ g" := (comp f g) : cat_scope.
Open Scope cat_scope.

Section Lemmas.
  Context {C : pcat}.
  Implicit Types (a b c d : C).

  Lemma idL a b (f : hom a b) : id ∘ f = f.
  Proof. by case: C a b f => /=T[]. Qed.

  Lemma idR a b (f : hom a b) : f ∘ id = f.
  Proof. by case: C a b f => /=T[]. Qed.

  Lemma compA a b c d (f : hom a b) (g : hom b c) (h : hom c d) :
    (f ∘ g) ∘ h = f ∘ (g ∘ h).
  Proof. by case: C a b c d f g h => /=T[]. Qed.

  Global Instance comp_proper a b c :
    Proper (ord_op ++> ord_op ++> ord_op) (@comp C a b c).
  Proof. by case: C a b c => /=T[]. Qed.
End Lemmas.

(** ** The 2-poset of posets *)

Section OrdtypeCat.
  Definition mfun_hom (A B : ordType) : ordType :=
    Eval hnf in [ordType of {mono A -> B}].

  Definition mfun_id (A : ordType) : {mono A -> A} :=
    @mfun _ _ idfun (fun _ _ => idfun).

  Program Definition mfun_comp (A B C : ordType) (f : {mono A -> B}) (g : {mono B -> C}) : {mono A -> C} :=
    mk_mfun (g \o f) _.
  Next Obligation. by rewrite H. Qed.

  Lemma mfun_idL (A B : ordType) (f : {mono A -> B}) :
    mfun_comp (mfun_id A) f = f.
  Proof. exact: mfun_eq. Qed.

  Lemma mfun_idR (A B : ordType) (f : {mono A -> B}) :
    mfun_comp f (mfun_id B) = f.
  Proof. exact: mfun_eq. Qed.

  Lemma mfun_compA (A B C D : ordType) (f : {mono A -> B}) (g : {mono B -> C}) (h : {mono C -> D}) :
    mfun_comp (mfun_comp f g) h = mfun_comp f (mfun_comp g h).
  Proof. exact: mfun_eq. Qed.

  Global Instance mfun_comp_proper (A B C : ordType) :
    Proper (ord_op ++> ord_op ++> ord_op) (@mfun_comp A B C).
  Proof.
    move=> f1 f2 l1 g1 g2 l2 x /=. by rewrite (l2 _) (l1 _).
  Qed.

  Canonical ordType_pcatMixin :=
    @Cat.Mixin ordType mfun_hom
               mfun_id mfun_comp mfun_idL mfun_idR mfun_compA mfun_comp_proper.
  Canonical ordType_pcat :=
    PCat ordType ordType_pcatMixin.
End OrdtypeCat.

(** ** Opposite Category *)

Section OpCat.
  Variable T : pcat.
  Definition op := T.

  Definition op_hom (a b : T) :=
    Eval hnf in [ordType of hom b a].

  Definition op_id (a : T) : op_hom a a := id.

  Definition op_comp (a b c : T) (f : op_hom a b) (g : op_hom b c) : op_hom a c :=
    g ∘ f.

  Definition op_idL (a b : T) (f : op_hom a b) : op_comp id f = f.
  Proof. exact: idR. Qed.

  Definition op_idR (a b : T) (f : op_hom a b) : op_comp f id = f.
  Proof. exact: idL. Qed.

  Definition op_compA (a b c d : T) (f : op_hom a b) (g : op_hom b c) (h : op_hom c d) :
    op_comp (op_comp f g) h = op_comp f (op_comp g h).
  Proof. unfold op_comp. symmetry. exact: compA. Qed.

  Global Instance op_comp_proper (a b c : T) :
    Proper (ord_op ++> ord_op ++> ord_op) (@op_comp a b c).
  Proof. move=> f1 f2 l1 g1 g2 l2. unfold op_comp. by rewrite l1 l2. Qed.

  Canonical op_pcatMixin :=
    @Cat.Mixin op op_hom op_id op_comp op_idL op_idR op_compA op_comp_proper.
  Canonical op_pcat :=
    PCat op op_pcatMixin.
End OpCat.

(** ** Functors *)

Polymorphic Cumulative Structure functor (A B : pcat) := mk_functor
  { omap :> A -> B
  ; map : forall a b, hom a b -> hom (omap a) (omap b)
  ; map1 : forall a, map (id : hom a a) = id
  ; mapC : forall a b c (f : hom a b) (g : hom b c), map f ∘ map g = map (f ∘ g)
  }.
Arguments mk_functor {A B} omap map map1 mapC.

Section FunctorOps.
  Implicit Types (A B C : pcat).

  Program Definition fun_id {A} : functor A A :=
    mk_functor (fun a => a) (fun a b f => f) _ _.

  Program Definition fun_comp {A B C} (F : functor A B) (G : functor B C) : functor A C :=
    mk_functor (fun a => G (F a)) (fun a b f => map G (map F f)) _ _.
  Next Obligation. by rewrite !map1. Qed.
  Next Obligation. etransitivity. apply: mapC. f_equal. exact: mapC. Qed.

  Program Definition fun_const {A B} (x : B) : functor A B :=
    mk_functor (fun _ => x) (fun _ _ _ => id) _ _.
  Next Obligation. exact: idR. Qed.
End FunctorOps.

(** ** Natural Transformations *)

Polymorphic Cumulative Structure nat {A B : pcat} (F G : functor A B) := mk_nat
  { components :> forall a, hom (F a) (G a)
  ; natP : forall a b (f : hom a b),
      map F f ∘ components _ = components _ ∘ map G f
  }.
Arguments mk_nat {A B F G} components natP.

Lemma nat_eq (A B : pcat) (F G : functor A B) (eta eps : nat F G) :
  (components eta = components eps) -> eta = eps.
Proof.
  destruct eta,eps => /= E. destruct E. f_equal. exact: pi.
Qed.

Section NatOrdType.
  Context {A B : pcat} (F G : functor A B).
  Implicit Types (eta eps psi : nat F G).

  Definition nat_le eta eps := forall x, eta x ≤ eps x.

  Lemma nat_le_refl : reflexive nat_le.
  Proof. move=> x. by []. Qed.

  Lemma nat_le_trans : transitive nat_le.
  Proof. move=> eps eta psi l1 l2 x. by transitivity (eta x). Qed.

  Canonical nat_proMixin := ProMixin nat_le_refl nat_le_trans.
  Canonical nat_proType := Eval hnf in ProType (nat F G) nat_proMixin.

  Lemma nat_le_eq : antisymmetric nat_le.
  Proof. move=> eta eps l1 l2. apply: nat_eq. fext=> x. exact: le_eq. Qed.

  Canonical nat_ordMixin := OrdMixin nat_le_eq.
  Canonical nat_ordType := Eval hnf in OrdType (nat F G) nat_ordMixin.
End NatOrdType.

(** ** Functor Category *)

Section FunCat.
  Context (A B : pcat).
  Implicit Types (F G H : functor A B).

  Program Definition id_nat F : nat F F :=
    mk_nat (fun x => id) _.
  Next Obligation. etransitivity. exact: idR. symmetry. exact: idL. Qed.

  Program Definition comp_nat F G H (eta : nat F G) (eps : nat G H) :
    nat F H := mk_nat (fun x => eta x ∘ eps x) _.
  Next Obligation. by rewrite->compA,<- natP,<-!compA,<-natP. Qed.

  Lemma nat_idL F G (eta : nat F G) : comp_nat (id_nat _) eta = eta.
  Proof. apply: nat_eq; fext => x /=. exact: idL. Qed.

  Lemma nat_idR F G (eta : nat F G) : comp_nat eta (id_nat _) = eta.
  Proof. apply: nat_eq; fext=> x /=. exact: idR. Qed.

  Lemma nat_compA F G H I (eta1 : nat F G) (eta2 : nat G H) (eta3 : nat H I) :
    comp_nat (comp_nat eta1 eta2) eta3 = comp_nat eta1 (comp_nat eta2 eta3).
  Proof. apply: nat_eq; fext=> x /=. exact: compA. Qed.

  Lemma nat_comp_proper F G H :
    Proper (ord_op ++> ord_op ++> ord_op) (@comp_nat F G H).
  Proof.
    move=> eta1 eta2 l1 eps1 eps2 l2 x /=. apply: comp_proper. exact: l1. exact: l2.
  Qed.

  Canonical fun_pcatMixin :=
    @Cat.Mixin (functor A B) (@nat_ordType A B) id_nat comp_nat nat_idL nat_idR nat_compA nat_comp_proper.
  Canonical fun_pcat :=
    PCat (functor A B) fun_pcatMixin.
End FunCat.

Set Printing Universes.
Check fun_pcat ordType_pcat ordType_pcat.

(*** old stuff ***)


Cumulative Structure isPCategory {Obj : Type} {Hom : Obj -> Obj -> ordType}
          (id : forall {x}, Hom x x)
          (comp : forall {x y z}, Hom y z -> Hom x y -> Hom x z) :=
  { _ : forall x y (f : Hom x y), comp id f = f
  ; _ : forall x y (f : Hom x y), comp f id = f
  ; _ : forall w x y z (f : Hom w x) (g : Hom x y) (h : Hom y z),
      comp (comp h g) f = comp h (comp g f)
  ; _ : forall x y z, Proper (ord_op ++> ord_op ++> ord_op) (@comp x y z)
  }.

Cumulative Structure PCategory :=
  { Obj  :> Type
  ; Hom  :> Obj -> Obj -> ordType
  ; id   : forall x, Hom x x
  ; comp : forall x y z, Hom y z -> Hom x y -> Hom x z
  ; category_is_category :> isPCategory id comp
  }.
Arguments Obj  {C} : rename.
Arguments Hom  {C} x y : rename.
Arguments id   {C x} : rename.
Arguments comp {C x y z} g f : rename.

Notation "f ∘ g" := (comp f g) : cat_scope.
Open Scope cat_scope.

Lemma idL {C : PCategory} {x y : C} (f : C x y) : id ∘ f = f.
Proof. by case: C x y f => /= T H id comp []. Qed.

Lemma idR {C : PCategory} {x y : C} (f : C x y) : f ∘ id = f.
Proof. by case: C x y f => /= T H id comp []. Qed.

Lemma compA {C : PCategory} {w x y z : C} (f : C w x) (g : C x y) (h : C y z) :
  (h ∘ g) ∘ f = h ∘ (g ∘ f).
Proof. by case: C w x y z f g h => /= T H id comp []. Qed.

Global Instance comp_proper {C : PCategory} {x y z : C} :
  Proper (ord_op ++> ord_op ++> ord_op) (@comp C x y z).
Proof. by case: C x y z => /= T H id comp []. Qed.

(** ** Functors *)

Cumulative Structure Functor (A B : PCategory) :=
  { omap :> A -> B
  ; fmap : forall {x y}, A x y -> B (omap x) (omap y)
  ; _    : forall {x}, fmap (id : A x x) = id
  ; _    : forall {x y z} (f : A x y) (g : A y z),
      fmap g ∘ fmap f = fmap (g ∘ f)
  }.
Arguments fmap {A B} F {x y} f : rename.

Lemma fmap_id {A B} (F : Functor A B) {x : A} :
  fmap F (id : A x x) = id.
Proof. by case: F. Qed.

Lemma fmap_comp {A B} (F : Functor A B) {x y z : A} (f : A x y) (g : A y z) :
  fmap F g ∘ fmap F f = fmap F (g ∘ f).
Proof. by case: F. Qed.

(** ** Lax Natural Transformations *)

Cumulative Structure LaxNat {A B : PCategory} (F G : Functor A B) :=
  { nat_eta :> forall x, B (F x) (G x)
  ; nat_is_natural : forall x y (f : A x y),
      nat_eta y ∘ fmap F f ≤ fmap G f ∘ nat_eta x
  }.

Lemma natP {A B : PCategory} {F G : Functor A B} (eta : LaxNat F G) {x y} (f : A x y) :
  eta y ∘ fmap F f ≤ fmap G f ∘ eta x.
Proof. by case: eta. Qed.

Lemma nat_eq {A B : PCategory} {F G : Functor A B} (eta eps : LaxNat F G) :
  (forall x, eta x = eps x) -> eta = eps.
Proof.
  move: eta eps => -[eta etaP] [eps epsP] /=.
  move=> H. apply fext in H. subst eta. f_equal. exact: pi.
Qed.

(** order on laxnat *)

Section LaxNatOrdType.
  Context {A B : PCategory} (F G : Functor A B).
  Implicit Types (eta eps psi : LaxNat F G).

  Definition lax_le eta eps := forall x, eta x ≤ eps x.

  Lemma lax_le_refl : reflexive lax_le.
  Proof. move=> x. by []. Qed.

  Lemma lax_le_trans : transitive lax_le.
  Proof. move=> eps eta psi l1 l2 x. by transitivity (eta x). Qed.

  Canonical lax_proMixin := ProMixin lax_le_refl lax_le_trans.
  Canonical lax_proType := Eval hnf in ProType (LaxNat F G) lax_proMixin.

  Lemma lax_le_eq : antisymmetric lax_le.
  Proof. move=> eta eps l1 l2. apply: nat_eq => x. exact: le_eq. Qed.

  Canonical lax_ordMixin := OrdMixin lax_le_eq.
  Canonical lax_ordType := Eval hnf in OrdType (LaxNat F G) lax_ordMixin.
End LaxNatOrdType.

Program Definition id_nat {A B : PCategory} (F : Functor A B) : LaxNat F F :=
  {| nat_eta := fun x => id |}.
Next Obligation. by rewrite idL idR. Qed.

Program Definition comp_nat {A B : PCategory} {F G H : Functor A B} (eps : LaxNat G H) (eta : LaxNat F G) :
  LaxNat F H := {| nat_eta := fun x => eps x ∘ eta x |}.
Next Obligation. by rewrite compA natP -compA natP compA. Qed.

Definition mfun_id (A : proType) : {mono A -> A} :=
  @mfun _ _ idfun (fun _ _ => idfun).

Program Definition mfun_comp (A B C : proType) (g : {mono B -> C}) (f : {mono A -> B}) : {mono A -> C} :=
  mk_mfun (g \o f) _.
Next Obligation. by rewrite H. Qed.

Global Instance mfun_comp_proper (A B C : proType) :
  Proper (ord_op ++> ord_op ++> ord_op) (@mfun_comp A B C).
Proof.
  move=> f1 f2 l1 g1 g2 l2 x /=. by rewrite (l1 (g1 x)) (l2 x).
Qed.

Lemma ordType_is_category : @isPCategory ordType (fun A B => [ordType of {mono A -> B}]) mfun_id mfun_comp.
Proof.
  constructor.
  - move=> A B /= f. exact: le_eq.
  - move=> A B /= f. exact: le_eq.
  - move=> A B C D /= f g h. exact: le_eq.
  - move=> A B C /=. exact _.
Qed.

Definition PSet : PCategory :=
  {| Obj  := ordType
   ; Hom  := fun A B => [ordType of {mono A -> B}]
   ; id   := mfun_id
   ; comp := mfun_comp
   ; category_is_category := ordType_is_category
  |}.

Lemma Fun_is_category {A B : PCategory} :
  @isPCategory (Functor A B) (fun F G => [ordType of LaxNat F G]) id_nat (@comp_nat A B).
Proof.
  constructor.
  - move=> F G f. apply: nat_eq => x /=. exact: idL.
  - move=> F G f. apply: nat_eq => x /=. exact: idR.
  - move=> F G H I f g h. apply: nat_eq => x /=. exact: compA.
  - move=> F G H /= e1 e2 l1 f1 f2 l2 x /=. by rewrite (l1 x) (l2 x).
Qed.

Polymorphic Definition Fun (A B : PCategory) :=
  {| Obj := Functor A B
   ; Hom := fun F G => [ordType of LaxNat F G]
   ; id  := id_nat
   ; comp := (@comp_nat A B)
   ; category_is_category := Fun_is_category
  |}.

Lemma Op_is_category (A : PCategory) :
  @isPCategory A (fun x y => A y x) (fun x => id) (fun x y z g f => f ∘ g).
Proof.
  constructor.
  - move=> x y f. exact: idR.
  - move=> x y f. exact: idL.
  - move=> w x y z f g h. by rewrite compA.
  - move=> x y z f1 f2 l1 g1 g2 l2. by rewrite l1 l2.
Qed.

Polymorphic Definition Op (A : PCategory) :=
  {| Obj := A
   ; Hom := fun x y => A y x
   ; id := fun x => id
   ; comp := fun x y z g f => f ∘ g
   ; category_is_category := Op_is_category A
  |}.

Polymorphic Definition LaxPresheaf (A : PCategory) : PCategory :=
  Fun (Op A) PSet.

(** ** The inverse of a natural equivalence is natural *)

Structure is_iso {A : PCategory} {x y : A} (f : Hom x y) :=
  { iso_inverse :> Hom y x
  ; iso_sect : iso_inverse ∘ f = id
  ; iso_retr : f ∘ iso_inverse = id
  }.

Structure Iso {A : PCategory} (x y : A) :=
  { iso_hom :> Hom x y
  ; iso_hom_is_iso :> is_iso iso_hom
  }.

Lemma is_prop_is_iso (A : PCategory) (x y : A) (f : Hom x y) :
  is_prop (is_iso f).
Proof.
  move=> [g1 s1 r1] [g2 s2 r2]. suff: g1 = g2. move=> E. destruct E. f_equal; exact: pi.
    by rewrite -[g1]idR -r2 -compA s1 idL.
Qed.

Lemma iso_eq {A : PCategory} {x y : A} (f g : Iso x y) :
  f = g :> Hom _ _ -> f = g.
Proof.
  destruct f,g => /= E. destruct E. f_equal. exact: is_prop_is_iso.
Qed.

(** Composition of Functors *)

Section FunctorCompose.
  Context {A B C : PCategory} (F : Functor A B) (G : Functor B C).

  Program Definition fun_compose : Functor A C :=
    @Build_Functor A C (fun x => G (F x)) (fun x y f => fmap G (fmap F f)) _ _.
  Next Obligation. by rewrite !fmap_id. Qed.
  Next Obligation. by rewrite !fmap_comp. Qed.
End FunctorCompose.

(** Constant Functor *)

Program Definition fun_const {A B : PCategory} (x : B) : Functor A B :=
  @Build_Functor A B (fun _ => x) (fun _ _ _ => id) _ _.
Next Obligation. exact: idR. Qed.

(** Natural transformations *)

Cumulative Structure Nat {A B : PCategory} (F G : Functor A B) :=
  { rnat_eta :> forall x, B (F x) (G x)
  ; rnat_is_natural : forall x y (f : A x y),
      rnat_eta y ∘ fmap F f = fmap G f ∘ rnat_eta x
  }.

Lemma rnatP {A B : PCategory} {F G : Functor A B} (eta : Nat F G) {x y} (f : A x y) :
  eta y ∘ fmap F f = fmap G f ∘ eta x.
Proof. by case: eta. Qed.

Lemma rnat_eq {A B : PCategory} {F G : Functor A B} (eta eps : Nat F G) :
  (forall x, eta x = eps x) -> eta = eps.
Proof.
  move: eta eps => -[eta etaP] [eps epsP] /=.
  move=> H. apply fext in H. subst eta. f_equal. exact: pi.
Qed.

(** order on nat *)

Section NatOrdType.
  Context {A B : PCategory} (F G : Functor A B).
  Implicit Types (eta eps psi : Nat F G).

  Definition nat_le eta eps := forall x, eta x ≤ eps x.

  Lemma nat_le_refl : reflexive nat_le.
  Proof. move=> x. by []. Qed.

  Lemma nat_le_trans : transitive nat_le.
  Proof. move=> eps eta psi l1 l2 x. by transitivity (eta x). Qed.

  Canonical nat_proMixin := ProMixin nat_le_refl nat_le_trans.
  Canonical nat_proType := Eval hnf in ProType (Nat F G) nat_proMixin.

  Lemma nat_le_eq : antisymmetric nat_le.
  Proof. move=> eta eps l1 l2. apply: rnat_eq => x. exact: le_eq. Qed.

  Canonical nat_ordMixin := OrdMixin nat_le_eq.
  Canonical nat_ordType := Eval hnf in OrdType (Nat F G) nat_ordMixin.
End NatOrdType.

Program Definition id_rnat {A B : PCategory} (F : Functor A B) : Nat F F :=
  {| rnat_eta := fun x => id |}.
Next Obligation. by rewrite idL idR. Qed.

Program Definition comp_rnat {A B : PCategory} {F G H : Functor A B} (eps : Nat G H) (eta : Nat F G) :
  Nat F H := {| rnat_eta := fun x => eps x ∘ eta x |}.
Next Obligation. by rewrite compA rnatP -compA rnatP compA. Qed.

(** Hom Functor *)

Program Definition fun_hom {A : PCategory} : Functor A (PreSheaf A).

(*Section NatEquiv.
  Context {A B : PCategory} {F G : Functor A B}.
  Variable (eta : forall x, Hom (F x) (G x)).
  Hypothesis nat : forall x y (f : Hom x y),
      eta y ∘ fmap F f = fmap G f ∘ eta x.

  (* eta is an equivalence *)
  Variable etai : forall x, Hom (G x) (F x).
  Hypothesis sect : forall x, etai x ∘ eta x = 1.
  Hypothesis retr : forall x, eta x ∘ etai x = 1.

  Lemma nat_inverse_nat x y (f : Hom x y) :
    etai y ∘ fmap G f = fmap F f ∘ etai x.
  Proof.
    rewrite -[fmap F _]idL -sect !compA.
      by rewrite -[_ ∘ (_ ∘ etai _)]compA nat compA retr idR.
  Qed.
End NatEquiv.*)

(** ** Limits *)

Structure is_product {C : PCategory} (x y z : C) :=
  { prod :> forall u, Hom u x * Hom u y -> Hom u z
  ; prod_is_equiv : forall u, is_equiv (@prod u)
  ; prod_is_nat   : forall u v (f : Hom u v) (g : Hom v x) (h : Hom v y),
      prod (g, h) ∘ f = prod (g ∘ f, h ∘ f)
  ; prod_is_mono  : forall u (f1 f2 : Hom u x) (g1 g2 : Hom u y),
      f1 ≤ f2 -> g1 ≤ g2 -> prod (f1, g1) ≤ prod (f2, g2)
  }.

Section ProductProperties.
  Variable (C : PCategory) (x y z : C).
  Hypothesis H : is_product x y z.

  Definition pi  : Hom z x := (prod_is_equiv H z id).1.
  Definition pi' : Hom z y := (prod_is_equiv H z id).2.

  Definition prd {u} (f : Hom u x) (g : Hom u y) : Hom u z :=
    H u (f, g).

  Lemma prd_pi {u} (f : Hom u x) (g : Hom u y) :
    pi ∘ prd f g = f.
  Proof.
    rewrite/pi/prd.

Definition is_product {C : PCategory} (x y z : C) :=
  forall u, (Hom u x * Hom u y) <~> Hom u z.

Section ProductProperties.
  Variable (C : PCategory) (x y z : C).
  Hypothesis H : is_product x y z.

  Definition pi  : Hom z x := ((H z)^-1 1).1.
  Definition pi' : Hom z y := ((H z)^-1 1).2.

  Definition prod {u} (f : Hom u x) (g : Hom u y) : Hom u z :=
    H u (f, g).

  Lemma prod_pi {u} (f : Hom u x) (g : Hom u y) :
    pi ∘ prod f g = f.
  Proof.
    rewrite/pi/prod.