Require Import base protype ordtype addr.

Inductive nmap (A : Type) :=
| tip (x : A)
| branchL (x : option A) (s : nmap A)
| branchR (x : option A) (s : nmap A)
| branch2 (x : option A) (s t : nmap A).

Inductive fmap (A : Type) :=
| empty
| nempty (s : nmap A).
Arguments empty {A}.
Coercion nempty : nmap >-> fmap.

Section Operations.
  Import BinPos.

  (** Lookup *)

  Fixpoint nmap_lookup {A : Type} (h : nmap A) (a : addr) : option A :=
    match h, a with
    | tip x, xH => Some x
    | branchL x _, xH => x
    | branchR x _, xH => x
    | branch2 x _ _, xH => x
    | branchL _ s, xO p => nmap_lookup s p
    | branchR _ s, xI p => nmap_lookup s p
    | branch2 _ s _, xO p => nmap_lookup s p
    | branch2 _ _ s, xI p => nmap_lookup s p
    | _, _ => None
    end.
  Coercion nmap_lookup : nmap >-> Funclass.

  Definition fmap_lookup {A : Type} (h : fmap A) (a : addr) : option A :=
    match h with
    | empty => None
    | nempty h => h a
    end.
  Coercion fmap_lookup : fmap >-> Funclass.
  (* Ambiguous but convertible coercions to Funclass *)

  Definition nmap_dom {A} (h : nmap A) (a : addr) : bool := h a.
  Canonical nmap_predType A := @mkPredType addr (nmap A) nmap_dom.

  Definition fmap_dom {A} (h : fmap A) (a : addr) : bool := h a.
  Canonical fmap_predType A := @mkPredType addr (fmap A) fmap_dom.

  (** Singletons *)

  Fixpoint singleton {A : Type} (a : addr) (x : A) : nmap A :=
    match a with
    | xH => tip x
    | xO p => branchL None (singleton p x)
    | xI p => branchR None (singleton p x)
    end.

  (** Disjointness *)

  Fixpoint ndisjoint {A} (h1 h2 : nmap A) : bool :=
    if h1 xH && h2 xH then false else
      match h1, h2 with
      | tip _, _ => true
      | _, tip _ => true
      | branchL _ _, branchR _ _ => true
      | branchR _ _, branchL _ _ => true
      | branchL _ h1, branchL _ h2 => ndisjoint h1 h2
      | branchL _ h1, branch2 _ h2 _ => ndisjoint h1 h2
      | branch2 _ h1 _, branchL _ h2 => ndisjoint h1 h2
      | branchR _ h1, branchR _ h2 => ndisjoint h1 h2
      | branchR _ h1, branch2 _ _ h2 => ndisjoint h1 h2
      | branch2 _ _ h1, branchR _ h2 => ndisjoint h1 h2
      | branch2 _ h11 h12, branch2 _ h21 h22 => ndisjoint h11 h21 && ndisjoint h12 h22
      end.

  Definition disjoint {A} (h1 h2 : fmap A) : bool :=
    match h1, h2 with
    | nempty h1, nempty h2 => ndisjoint h1 h2
    | _, _ => true
    end.

  (** Union *)

  Definition ounion {A : Type} (x y : option A) : option A :=
    if x is None then y else x.

  Fixpoint nunion {A : Type} (h1 h2 : nmap A) : nmap A :=
    match h1, h2 with
    | tip x, tip _ => tip x
    | branchL x h1, tip y => branchL (ounion x (Some y)) h1
    | branchR x h1, tip y => branchR (ounion x (Some y)) h1
    | branch2 x h11 h12, tip y => branch2 (ounion x (Some y)) h11 h12
    | tip x, branchL _ h2 => branchL (Some x) h2
    | tip x, branchR _ h2 => branchR (Some x) h2
    | tip x, branch2 _ h21 h22 => branch2 (Some x) h21 h22
    | branchL x h1, branchL y h2 => branchL (ounion x y) (nunion h1 h2)
    | branchL x h1, branchR y h2 => branch2 (ounion x y) h1 h2
    | branchL x h1, branch2 y h21 h22 => branch2 (ounion x y) (nunion h1 h21) h22
    | branchR x h1, branchL y h2 => branch2 (ounion x y) h2 h1
    | branchR x h1, branchR y h2 => branchR (ounion x y) (nunion h1 h2)
    | branchR x h1, branch2 y h21 h22 => branch2 (ounion x y) h21 (nunion h1 h22)
    | branch2 x h11 h12, branchL y h2 => branch2 (ounion x y) (nunion h11 h2) h12
    | branch2 x h11 h12, branchR y h2 => branch2 (ounion x y) h11 (nunion h12 h2)
    | branch2 x h11 h12, branch2 y h21 h22 =>
      branch2 (ounion x y) (nunion h11 h21) (nunion h12 h22)
    end.

  Definition union {A} (h1 h2 : fmap A) : fmap A :=
    match h1, h2 with
    | empty, _ => h2
    | _, empty => h1
    | nempty h1, nempty h2 => nempty (nunion h1 h2)
    end.

  (** Difference *)

  Definition odiff {A B : Type} (x : option A) (y : option B) : option A :=
    if y then None else x.

  Definition branch {A} (x : option A) (h1 h2 : fmap A) : fmap A :=
    match x, h1, h2 with
    | None, empty, empty => empty
    | Some x, empty, empty => tip x
    | _, nempty h1, empty => branchL x h1
    | _, empty, nempty h2 => branchR x h2
    | _, nempty h1, nempty h2 => branch2 x h1 h2
    end.

  Definition fbranchL {A} (x : option A) (h : fmap A) : fmap A :=
    branch x h empty.

  Definition fbranchR {A} (x : option A) (h : fmap A) : fmap A :=
    branch x empty h.

  Fixpoint ndiff {A B : Type} (h1 : nmap A) (h2 : nmap B) : fmap A :=
    match h1, h2 with
    | tip y, branchL None _ => nempty (tip y)
    | tip y, branchR None _ => nempty (tip y)
    | tip y, branch2 None _ _ => nempty (tip y)
    | tip _, _ => empty
    | branchL _ h, tip _ => nempty (branchL None h)
    | branchR _ h, tip _ => nempty (branchR None h)
    | branch2 _ h1 h2, tip _ => nempty (branch2 None h1 h2)
    | branchL x h1, branchL y h2 => fbranchL (odiff x y) (ndiff h1 h2)
    | branchL x h1, branchR y h2 => nempty (branchL (odiff x y) h1)
    | branchL x h1, branch2 y h21 h22 => fbranchL (odiff x y) (ndiff h1 h21)
    | branchR x h1, branchL y _ => nempty (branchR (odiff x y) h1)
    | branchR x h1, branchR y h2 => fbranchR (odiff x y) (ndiff h1 h2)
    | branchR x h1, branch2 y h21 h22 => fbranchR (odiff x y) (ndiff h1 h22)
    | branch2 x h11 h12, branchL y h2 => branch (odiff x y) (ndiff h11 h2) h12
    | branch2 x h11 h12, branchR y h2 => branch (odiff x y) h11 (ndiff h12 h2)
    | branch2 x h11 h12, branch2 y h21 h22 =>
      branch (odiff x y) (ndiff h11 h21) (ndiff h12 h22)
    end.

  Definition diff {A B} (h1 : fmap A) (h2 : fmap B) : fmap A :=
    match h1, h2 with
    | _, empty => h1
    | empty, _ => empty
    | nempty h1, nempty h2 => ndiff h1 h2
    end.

  (** Lemmas *)

  Lemma dom_nmapE {A} (a : addr) (h : nmap A) : (a \in h) = (h a).
  Proof. by []. Qed.

  Lemma dom_fmapE {A} (a : addr) (h : fmap A) : (a \in h) = (h a).
  Proof. by []. Qed.


  Lemma lookup_empty {A} (a : addr) :
    (@empty A) a = None.
  Proof. by []. Qed.

  Lemma dom_empty {A} (a : addr) :
    a \notin (@empty A).
  Proof. by []. Qed.


  Lemma lookup_singleton {A} (a1 a2 : addr) (x : A) :
    singleton a1 x a2 = (if a1 == a2 then Some x else None).
  Proof.
    elim: a1 a2 => [a1 ih|a1 ih|] [a2|a2|] //=; by rewrite ih eqE.
  Qed.

  Lemma dom_singleton {A} (a1 a2 : addr) (x : A) :
    (a2 \in singleton a1 x) = (a1 == a2).
  Proof.
    rewrite dom_nmapE lookup_singleton. by case: ifP.
  Qed.


  Lemma match_some_eta {A} (x : option A) :
    match x with Some x => Some x | None => None end = x.
  Proof. by case: x. Qed.

  Lemma lookup_nunion {A} (h1 h2 : nmap A) (a : addr) :
    nunion h1 h2 a = (if h1 a is Some x then Some x else h2 a).
  Proof.
    elim: h1 h2 a => [x|[x|] h1 ih|[x|] h1 ih|[x|] h11 ih1 h12 ih2]
      [y|y h2|y h2|y h21 h22] [a|a|] //=; by rewrite match_some_eta.
  Qed.

  Lemma lookup_union {A} (h1 h2 : fmap A) (a : addr) :
    union h1 h2 a = (if a \in h1 then h1 a else h2 a).
  Proof.
    move: h1 h2 => -[|h1] [|h2] //=.
    - rewrite dom_fmapE/=. by case: (h1 a).
    - rewrite dom_fmapE lookup_nunion /=. by case: (h1 a).
  Qed.

  Lemma dom_union {A} (h1 h2 : fmap A) (a : addr) :
    (a \in union h1 h2) = (a \in h1) || (a \in h2).
  Proof.
    rewrite !dom_fmapE lookup_union dom_fmapE. by case: (h1 a).
  Qed.


  Lemma match_some_const {A B} (x : option A) (y : B) :
    match x with Some _ => y | None => y end = y.
  Proof. by case: x. Qed.

  Lemma fmap_lookup_branch {A} (x : option A) (h1 h2 : fmap A) (a : addr) :
    branch x h1 h2 a =
      match a with
      | xH => x
      | xO a => h1 a
      | xI a => h2 a
      end.
  Proof. by move: x h1 h2 a => -[x|] [|h1] [|h2] [a|a|]. Qed.

  Lemma lookup_ndiff {A B} (h1 : nmap A) (h2 : nmap B) (a : addr) :
    ndiff h1 h2 a = (if h2 a is Some x then None else h1 a).
  Proof.
    elim: h1 h2 a => [x|x h1 ih|x h1 ih|x h11 ih1 h12 ih2]
      [y|y h2|y h2|y h21 h22] [a|a|] //=;
      try destruct x; rewrite/fbranchL/fbranchR ?fmap_lookup_branch//=;
      try rewrite ?match_some_const //; try case: y => //=.
  Qed.

  Lemma lookup_diff {A B} (h1 : fmap A) (h2 : fmap B) (a : addr) :
    diff h1 h2 a = if a \in h2 then None else h1 a.
  Proof.
    move: h1 h2 => [|h1] [|h2] //=. by case: ifP.
    rewrite lookup_ndiff dom_fmapE /=. by case: (h2 a).
  Qed.

  Lemma dom_diff {A B} (h1 : fmap A) (h2 : fmap B) (a : addr) :
    a \in diff h1 h2 = (a \in h1) && (a \notin h2).
  Proof.
    rewrite !dom_fmapE/= lookup_diff dom_fmapE/=.
    case: (h2 a) => //= [_|]. by rewrite andbF. by rewrite andbT.
  Qed.


  Lemma ndisjointP {A} (h1 h2 : nmap A) :
    reflect (forall a, ~~h1 a || ~~h2 a) (ndisjoint h1 h2).
  Proof.
    apply: (iffP idP).
    - elim: h1 h2 => [x|[x|] h1 ih|[x|] h1 ih|[x|] h11 ih1 h12 ih2]
        [y|[y|] h2|[y|] h2|[y|] h21 h22] //= h [a|a|] //=;
        rewrite ?orbT //; try exact: ih; try exact: ih1; try exact: ih2;
        move: h => /andP[h1 h2]; try exact: ih2; exact: ih1.
    - elim: h1 h2 => [x|[x|] h1 ih|[x|] h1 ih|[x|] h11 ih1 h12 ih2]
        [y|[y|] h2|[y|] h2|[y|] h21 h22] //= h; try exact: (h xH);
        try (apply: ih => a; try exact: (h (xO a)); exact: (h (xI a)));
        try (apply/andP; split);
        try (apply: ih1 => a; try exact: (h (xO a)); exact: (h (xI a)));
        try (apply: ih2 => a; try exact: (h (xO a)); exact: (h (xI a))).
  Qed.

  Lemma disjointP {A} (h1 h2 : fmap A) :
    reflect (forall a, (a \notin h1) || (a \notin h2)) (disjoint h1 h2).
  Proof.
    move: h1 h2 => -[|h1] [|h2] /=; try exact: (iffP idP).
    apply: (iffP idP) => [_ a|//]; by rewrite orbT.
    exact: ndisjointP.
  Qed.

  Lemma disjointC {A} (h1 h2 : fmap A) :
    disjoint h1 h2 = disjoint h2 h1.
  Proof.
    apply: (sameP idP). apply: (iffP idP) => /disjointP h;
    apply/disjointP => a; rewrite orbC; exact: h.
  Qed.

  Lemma disjointhU {A} (h1 h2 h3 : fmap A) :
    disjoint h1 (union h2 h3) = (disjoint h1 h2 && disjoint h1 h3).
  Proof.
    apply: (sameP idP). apply: (iffP idP).
    - move=>/andP[/disjointP p /disjointP q].
      apply/disjointP => a. by rewrite dom_union negb_or orb_andr p q.
    - move=>/disjointP p. apply/andP. split; apply/disjointP => a;
      move: (p a); rewrite dom_union negb_or orb_andr => /andP[]//.
  Qed.

  Lemma heap_disjointUh {A} (h1 h2 h3 : fmap A) :
    disjoint (union h1 h2) h3 = (disjoint h1 h3 && disjoint h2 h3).
  Proof.
    rewrite disjointC disjointhU. congr andb; exact: disjointC.
  Qed.

  Lemma disjoint_diff {A} (h1 h2 : fmap A) :
    disjoint h1 (diff h2 h1).
  Proof.
    apply/disjointP => a. rewrite dom_diff negb_and.
    case: (a \notin h1) => //=. by rewrite orbT.
  Qed.

  (** Extensionality *)

  Lemma nmap_nonempty {A} (h : nmap A) :
    exists a, h a.
  Proof.
    elim: h => [x|x h [a ih]|x h [a ih]|x h1 [a ih] h2 _].
      by exists xH. by exists (xO a). by exists (xI a). by exists (xO a).
  Qed.

  Lemma nmap_ext {A} (h1 h2 : nmap A) :
    h1 =1 h2 -> h1 = h2.
  Proof.
    elim: h1 h2 => [x|x h1 ih|x h1 ih|x h11 ih1 h12 ih2] [y|y h2|y h2|y h21 h22] //= h.
    - by move: (h xH) => [->].
    - case: (nmap_nonempty h2) => a. move: (h (xO a)). by case: (h2 a).
    - case: (nmap_nonempty h2) => a. move: (h (xI a)). by case: (h2 a).
    - case: (nmap_nonempty h21) => a. move: (h (xO a)). by case: (h21 a).
    - case: (nmap_nonempty h1) => a. move: (h (xO a)). by case: (h1 a).
    - congr branchL. exact: (h xH). apply: ih => a. exact: (h (xO a)).
    - case: (nmap_nonempty h1) => a. move: (h (xO a)). by case: (h1 a).
    - case: (nmap_nonempty h22) => a. move: (h (xI a)). by case: (h22 a).
    - case: (nmap_nonempty h1) => a. move: (h (xI a)). by case: (h1 a).
    - case: (nmap_nonempty h1) => a. move: (h (xI a)). by case: (h1 a).
    - congr branchR. exact: (h xH). apply: ih => a. exact: (h (xI a)).
    - case: (nmap_nonempty h21) => a. move: (h (xO a)). by case: (h21 a).
    - case: (nmap_nonempty h12) => a. move: (h (xI a)). by case: (h12 a).
    - case: (nmap_nonempty h12) => a. move: (h (xI a)). by case: (h12 a).
    - case: (nmap_nonempty h11) => a. move: (h (xO a)). by case: (h11 a).
    - congr branch2. exact: (h xH). apply: ih1 => a. exact: (h (xO a)).
      apply: ih2 => a. exact: (h (xI a)).
  Qed.

  Lemma fmap_ext {A} (h1 h2 : fmap A) :
    h1 =1 h2 -> h1 = h2.
  Proof.
    move: h1 h2 => -[|h1] [|h2] //= eqn.
    - case: (nmap_nonempty h2) => a. by move: (eqn a) => /= <-.
    - case: (nmap_nonempty h1) => a. by move: (eqn a) => /= ->.
    - congr nempty. apply: nmap_ext => a. exact: eqn.
  Qed.

  (** Monoid structure *)

  Lemma union0h {A} (h : fmap A) :
    union empty h = h.
  Proof. by []. Qed.

  Lemma unionh0 {A} (h : fmap A) :
    union h empty = h.
  Proof.
    apply: fmap_ext => a. rewrite lookup_union /= dom_fmapE.
      by case: (h a).
  Qed.

  Lemma heap_unionA {A} (h1 h2 h3 : fmap A) :
    union (union h1 h2) h3 = union h1 (union h2 h3).
  Proof.
    apply: fmap_ext => a. rewrite !lookup_union dom_union.
    rewrite dom_fmapE/=. by case: (h1 a).
  Qed.

  Lemma unionC {A} (h1 h2 : fmap A) :
    disjoint h1 h2 -> union h1 h2 = union h2 h1.
  Proof.
    move=> /disjointP dP. apply: fmap_ext => a. rewrite !lookup_union.
    move: (dP a). rewrite !dom_fmapE/=.
    case: (h1 a) => //=[x|]; by case: (h2 a).
  Qed.

  (** Cofiniteness *)

  Fixpoint nmap_pick_fresh {A} (h : nmap A) : addr :=
    match h with
    | tip _ => xO xH
    | branchL _ _ => xI xH
    | branchR _ _ => xO xH
    | branch2 _ h _ => xO (nmap_pick_fresh h)
    end.

  Definition pick_fresh {A} (h : fmap A) : addr :=
    match h with
    | empty => xH
    | nempty h => nmap_pick_fresh h
    end.

  Lemma dom_nmap_pick_fresh {A} (h : nmap A) :
    ~~(h (nmap_pick_fresh h)).
  Proof. by elim: h. Qed.

  Lemma dom_pick_fresh {A} (h : fmap A) :
    pick_fresh h \notin h.
  Proof.
    case: h => //= h. rewrite dom_fmapE/=. exact: dom_nmap_pick_fresh.
  Qed.
End Operations.

(** ** Partial Order on Heaps *)

Section Order.
  Variable (A : Type).

  Definition fmap_le (h1 h2 : fmap A) : Prop :=
    forall a x, h1 a = Some x -> h2 a = Some x.

  Lemma fmap_le_refl : reflexive fmap_le.
  Proof. firstorder. Qed.

  Lemma fmap_le_trans : transitive fmap_le.
  Proof. firstorder. Qed.

  Canonical fmap_proMixin := ProMixin fmap_le_refl fmap_le_trans.
  Canonical fmap_proType := Eval hnf in ProType (fmap A) fmap_proMixin.

  Lemma fmap_le_eq (h1 h2 : fmap A) :
    h1 ≤ h2 -> h2 ≤ h1 -> h1 = h2.
  Proof.
    move=> le1 le2. apply: fmap_ext => a.
    case e1:(h1 a) => [x|]. by move: e1 => /le1->.
    case e2:(h2 a) => [x|] //. move: e2 => /le2. by rewrite e1.
  Qed.

  Canonical fmap_ordMixin := OrdMixin fmap_le_eq.
  Canonical fmap_ordType := Eval hnf in OrdType (fmap A) fmap_ordMixin.

  Lemma fmap_le_dom (h1 h2 : fmap A) a :
    h1 ≤ h2 -> a \in h1 -> a \in h2.
  Proof.
    move=> le_h. rewrite !dom_fmapE. case e:(h1 a) => [x|] // _.
      by move: e => /le_h->.
  Qed.

  Lemma fmap_le_unionL (h1 h2 : fmap A) :
    h1 ≤ union h1 h2.
  Proof.
    move=> a x p. by rewrite lookup_union ifT // dom_fmapE p.
  Qed.

  Lemma fmap_le_unionR (h1 h2 : fmap A) :
    disjoint h1 h2 -> h2 ≤ union h1 h2.
  Proof.
    move=> p a x q. rewrite lookup_union ifF //.
    move/disjointP: p => /(_ a). rewrite !dom_fmapE q orbF.
      by case: (h1 a).
  Qed.

  Lemma fmap_le_unionE (h1 h2 h3 : fmap A) :
    h1 ≤ h3 -> h2 ≤ h3 -> union h1 h2 ≤ h3.
  Proof.
    move=> le1 le2 a x. rewrite lookup_union. case: ifP => _.
    exact: le1. exact: le2.
  Qed.

  Lemma fmap_leP (h1 h2 : fmap A) :
    (h1 ≤ h2) <-> union h1 (diff h2 h1) = h2.
  Proof.
    split.
    - move=> le_h. apply: le_eq => a x; rewrite lookup_union lookup_diff.
      + case: ifP. move=> _. exact: le_h. by move->.
      + case: ifP. rewrite dom_fmapE. case e:(h1 a) => [y|//] _.
          by move: e => /le_h -> ->. by move->.
    - move<-. exact: fmap_le_unionL.
  Qed.

  Lemma fmap_le_monoid (h1 h2 : fmap A) :
    (h1 ≤ h2) <-> exists h3, union h1 h3 = h2.
  Proof.
    split. move=> /fmap_leP eqn; eauto. move=> [h3 eqn].
    subst. exact: fmap_le_unionL.
  Qed.
End Order.
