Require Import base ord data.fintype ars.

Inductive sort :=
| prop
| type (n : nat).

Definition nat_le (m n : nat) : Prop := (m <= n)%bool.

Lemma nat_le_refl n : nat_le n n.
Proof. exact: leqnn. Qed.

Lemma nat_le_trans : transitive nat_le.
Proof. move=> x y z. exact: leq_trans. Qed.

Lemma nat_le_eq m n : nat_le m n -> nat_le n m -> m = n.
Proof.
  move=> le_m_n le_n_m. apply/eqP. by rewrite eqn_leq le_m_n le_n_m.
Qed.

Canonical nat_proMixin := ProMixin nat_le_refl nat_le_trans.
Canonical nat_proType := Eval hnf in ProType nat nat_proMixin.
Canonical nat_ordMixin := OrdMixin nat_le_eq.
Canonical nat_ordType := Eval hnf in OrdType nat nat_ordMixin.

Lemma nat_le_E (m n : nat) : (m ≤ n) = (m <= n).
Proof. by []. Qed.

Definition sort_le (m n : sort) : Prop :=
  match m, n with
  | prop, _ => True
  | _, prop => False
  | type m, type n => m ≤ n
  end.

Lemma sort_le_refl n : sort_le n n.
Proof. by case n=> /=. Qed.

Lemma sort_le_trans : transitive sort_le.
Proof. move=> []//m[]//=n[]//=k. exact: le_trans. Qed.

Lemma sort_le_eq m n : sort_le m n -> sort_le n m -> m = n.
Proof.
  destruct m, n => //= l1 l2. apply: ap. exact: le_eq.
Qed.

Canonical sort_proMixin := ProMixin sort_le_refl sort_le_trans.
Canonical sort_proType := Eval hnf in ProType sort sort_proMixin.
Canonical sort_ordMixin := OrdMixin sort_le_eq.
Canonical sort_ordType := Eval hnf in OrdType sort sort_ordMixin.

Definition sort_prod (m n : sort) : sort :=
  match m, n with
  | _, prop => prop
  | prop, _ => n
  | type m, type n => type (maxn m n)
  end.

Lemma sort_le_typeE m n :
  (type m ≤ type n) = (m <= n).
Proof. by []. Qed.

Global Instance sort_prod_proper :
  Proper (ord_op ++> ord_op ++> ord_op) sort_prod.
Proof.
  move=> [|m1] [|m2]//= l1 [|n1] [|n2] //= l2 /=.
  - apply: le_trans l2 _. rewrite sort_le_typeE. exact: leq_maxr.
  - rewrite sort_le_typeE geq_max. apply/andP; split.
    apply: leq_trans l1 _. exact: leq_maxl.
    apply: leq_trans l2 _. exact: leq_maxr.
Qed.

Definition sort_succ (m : sort) : sort :=
  match m with
  | prop => type 0
  | type m => type m.+1
  end.

(** **** Curry-style type theory with a hierarchy of predicative universes. *)

Inductive tm (n : nat) : Type :=
| var (x : fin n)
| univ (s : sort)
| app  (s t : tm n)
| lam  (s : tm n.+1)
| prod (s : tm n) (t : tm n.+1).

Fixpoint rename {m n} (f : ren m n) (s : tm m) : tm n :=
  match s with
  | var x => var (f x)
  | univ s => univ n s
  | app s t => app (rename f s) (rename f t)
  | lam s => lam (rename (up f) s)
  | prod s t => prod (rename f s) (rename (up f) t)
  end.

Definition subst m n := fin m -> tm n.

Definition upi {m n} (σ : subst m n) : subst m.+1 n.+1 :=
  var bound .: σ >> rename shift.

Fixpoint inst {m n} (f : subst m n) (s : tm m) : tm n :=
  match s with
  | var x => f x
  | univ s => univ n s
  | app s t => app (inst f s) (inst f t)
  | lam s => lam (inst (upi f) s)
  | prod s t => prod (inst f s) (inst (upi f) t)
  end.

(** ** Identity renaming law *)

Lemma upi_ids n : upi (@var n)  = @var n.+1.
Proof. fext=>/=. by case. Qed.

Lemma ren_id n (s : tm n) : rename id s = s.
Proof.
  elim: s => {n} //=.
  - by move=> n s-> t->.
  - move=> n s e. unfold up. fsimpl. by rewrite e.
  - move=> n s -> t e. unfold up. fsimpl. by rewrite e.
Qed.

(** ** Renaming as a special case of instantiation *)

Lemma ren_inst m n (f : ren m n) (s : tm m) :
  rename f s = inst (f >> @var _) s.
Proof.
  elim: s n f => //={m}[m s ihs t iht|m s ihs|m s ihs t iht] n f.
  - by rewrite ihs iht.
  - rewrite ihs. by fsimpl.
  - rewrite ihs iht. by fsimpl.
Qed.

(** ** Identity substitution law *)

Lemma inst_ids n (s : tm n) :
  inst (@var _) s = s.
Proof.
  by rewrite -ren_inst ren_id.
Qed.

(** ** Composition law *)

Lemma ren_inst_comp m n k (f : ren m n) (g : subst n k) (s : tm m) :
  inst g (rename f s) = inst (f >> g) s.
Proof.
  elim: s n k f g => //={m}[m s ihs t iht|m s ihs|m s ihs t iht] n k f g.
  - by rewrite ihs iht.
  - rewrite ihs. by fsimpl.
  - rewrite ihs iht. by fsimpl.
Qed.

Lemma ren_comp m n k (f : ren m n) (g : ren n k) (s : tm m) :
  rename g (rename f s) = rename (f >> g) s.
Proof.
  by rewrite ren_inst ren_inst_comp ren_inst.
Qed.

Lemma upi_up_comp m n k (f : subst m n) (g : ren n k) :
  upi f >> rename (up g) = upi (f >> rename g).
Proof.
  fext=> /=-[i|]//=. by rewrite !ren_comp.
Qed.

Lemma inst_ren_comp m n k (f : subst m n) (g : ren n k) (s : tm m) :
  rename g (inst f s) = inst (f >> rename g) s.
Proof.
  elim: s n k f g => //={m}[m s ihs t iht|m s ihs|m s ihs t iht] n k f g.
  - by rewrite ihs iht.
  - by rewrite ihs upi_up_comp.
  - by rewrite ihs iht upi_up_comp.
Qed.

Lemma upi_comp m n k (f : subst m n) (g : subst n k) :
  upi f >> inst (upi g) = upi (f >> inst g).
Proof.
  fext=> /=-[i|]//=. by rewrite ren_inst_comp inst_ren_comp.
Qed.

Lemma inst_comp m n k (f : subst m n) (g : subst n k) (s : tm m) :
  inst g (inst f s) = inst (f >> inst g) s.
Proof.
  elim: s n k f g => //={m}[m s ihs t iht|m s ihs|m s ihs t iht] n k f g.
  - by rewrite ihs iht.
  - by rewrite ihs upi_comp.
  - by rewrite ihs iht upi_comp.
Qed.

(** **** One-Step Reduction *)

Inductive step {m} : tm m -> tm m -> Prop :=
| step_beta s t u :
    u = inst (t .: @var m) s -> step (app (lam s) t) u
| step_appL s1 s2 t :
    step s1 s2 -> step (app s1 t) (app s2 t)
| step_appR s t1 t2 :
    step t1 t2 -> step (app s t1) (app s t2)
| step_lam s1 s2 :
    @step m.+1 s1 s2 -> step (lam s1) (lam s2)
| step_prodL A1 A2 B :
    step A1 A2 -> step (prod A1 B) (prod A2 B)
| step_prodR A B1 B2 :
    @step m.+1 B1 B2 -> step (prod A B1) (prod A B2).

Notation red := (star (@step _)).
Notation "s === t" := (conv (@step _) s t) (at level 50).

Definition sred {m n} (sigma tau : subst m n) :=
  forall x, red (sigma x) (tau x).

Lemma step_subst m n (sigma : subst m n) s t :
  step s t -> step (inst sigma s) (inst sigma t).
Proof.
  move=> st. elim: st n sigma => //={m s t}; eauto using @step.
  move=> m s t u -> n f; apply: step_beta. (* by autosubst. *)
Admitted.

(** **** Many-Step Reduction *)

Lemma red_app m (s1 s2 t1 t2 : tm m) :
  red s1 s2 -> red t1 t2 -> red (app s1 t1) (app s2 t2).
Proof.
  move=> A B. apply: (star_trans (app s2 t1)).
  - apply: (star_hom (@app _^~ t1)) A => x y. exact: step_appL.
  - apply: star_hom B => x y. exact: step_appR.
Qed.

Lemma red_lam m (s1 s2 : tm m.+1) :
  red s1 s2 -> red (lam s1) (lam s2).
Proof. apply: star_hom => x y. exact: step_lam. Qed.

Lemma red_prod m (s1 s2 : tm m) t1 t2 :
  red s1 s2 -> red t1 t2 -> red (prod s1 t1) (prod s2 t2).
Proof.
  move=> A B. apply: (star_trans (prod s2 t1)).
  - apply: (star_hom (@prod _^~ t1)) A => x y. exact: step_prodL.
  - apply: star_hom B => x y. exact: step_prodR.
Qed.

Lemma red_subst m n (f : subst m n) s t :
  red s t -> red (inst f s) (inst f t).
Proof. apply: star_hom. exact: step_subst. Qed.

Lemma sred_up m n (f g : subst m n) :
  sred f g -> sred (upi f) (upi g).
Proof.
  move=> A. hnf=>/=-[i|] //=. rewrite !ren_inst. apply: red_subst. exact: A.
Qed.

Hint Resolve @red_app @red_lam @red_prod @sred_up : red_congr.

Lemma red_compat m n (sigma tau : subst m n) s :
  sred sigma tau -> red (inst sigma s) (inst tau s).
Proof. elim: s n sigma tau =>/= *; eauto with red_congr. Qed.

(** **** Conversion *)

Definition sconv {m n} (sigma tau : subst m n) :=
  forall x, sigma x === tau x.

Lemma conv_app m (s1 s2 t1 t2 : tm m) :
  s1 === s2 -> t1 === t2 -> app s1 t1 === app s2 t2.
Proof.
  move=> A B. apply: (conv_trans (app s2 t1)).
  - apply: (conv_hom (@app _^~ t1)) A => x y. exact: step_appL.
  - apply: conv_hom B => x y. exact: step_appR.
Qed.

Lemma conv_lam m (s1 s2 : tm m.+1) :
  s1 === s2 -> lam s1 === lam s2.
Proof. apply: conv_hom => x y. exact: step_lam. Qed.

Lemma conv_prod m (s1 s2 : tm m) t1 t2 :
  s1 === s2 -> t1 === t2 -> prod s1 t1 === prod s2 t2.
Proof.
  move=> A B. apply: (conv_trans (prod s2 t1)).
  - apply: (conv_hom (@prod _^~ t1)) A => x y. exact: step_prodL.
  - apply: conv_hom B => x y. exact: step_prodR.
Qed.

Lemma conv_subst m n (sigma : subst m n) s t :
  s === t -> inst sigma s === inst sigma t.
Proof. apply: conv_hom. exact: step_subst. Qed.

Lemma sconv_up m n (sigma tau : subst m n) :
  sconv sigma tau -> sconv (upi sigma) (upi tau).
Proof.
  move=> A. hnf=>/=-[i|]//=. rewrite !ren_inst. exact: conv_subst.
Qed.

Lemma conv_compat m n (sigma tau : subst m n) s :
  sconv sigma tau -> inst sigma s === inst tau s.
Proof.
  elim: s n sigma tau => *; eauto using
    @conv_app, @conv_lam, @conv_prod, @sconv_up.
Qed.

Lemma conv_beta m (s : tm m.+1) t1 t2 :
  t1 === t2 -> inst (t1 .: @var _) s === inst (t2 .: @var _) s.
Proof. move=> c. by apply: conv_compat => -[]. Qed.

(** **** Church-Rosser theorem *)

Inductive pstep {n} : tm n -> tm n -> Prop :=
| pstep_beta s1 s2 t1 t2 u :
    u = inst (t2 .: @var _) s2 ->
    @pstep n.+1 s1 s2 -> pstep t1 t2 -> pstep (app (lam s1) t1) u
| pstep_var x : pstep (var x) (var x)
| pstep_sort u : pstep (univ _ u) (univ _ u)
| pstep_app s1 s2 t1 t2 :
    pstep s1 s2 -> pstep t1 t2 -> pstep (app s1 t1) (app s2 t2)
| pstep_lam s1 s2 :
    @pstep n.+1 s1 s2 -> pstep (lam s1) (lam s2)
| pstep_prod s1 s2 t1 t2 :
    pstep s1 s2 -> @pstep n.+1 t1 t2 -> pstep (prod s1 t1) (prod s2 t2).

Definition psstep {m n} (sigma tau : subst m n) :=
  forall x, pstep (sigma x) (tau x).

Fixpoint rho {n} (s : tm n) : tm n :=
  match s with
    | app (lam s) t => inst (rho t .: @var _) (rho s)
    | app s t => app (rho s) (rho t)
    | lam s => lam (rho s)
    | prod A B => prod (rho A) (rho B)
    | x => x
  end.

Lemma pstep_refl n (s : tm n) : pstep s s.
Proof. elim: s; eauto using @pstep. Qed.
Hint Resolve @pstep_refl.

Lemma step_pstep n (s t : tm n) : step s t -> pstep s t.
Proof. elim; eauto using @pstep. Qed.

Lemma pstep_red n (s t : tm n) : pstep s t -> red s t.
Proof.
  elim=> {n s t} //=; eauto with red_congr.
  move=> n s1 s2 t1 t2 u -> {u} _ A _ B. eapply starES. by econstructor.
  apply: (star_trans (inst (t1.:@var _) s2)). exact: red_subst.
  by apply: red_compat => -[|].
Qed.

Lemma pstep_subst m n (sigma : subst m n) s t :
  pstep s t -> pstep (inst sigma s) (inst sigma t).
Proof.
  move=> A. elim: A n sigma => /={m s t}; eauto using @pstep.
  move=> m s1 s2 t1 t2 u -> _ A _ B n sigma. eapply pstep_beta; eauto.
    (*by autosubst.*)
Admitted.

Lemma psstep_up m n (sigma tau : subst m n) :
  psstep sigma tau -> psstep (upi sigma) (upi tau).
Proof.
  move=> A. hnf=>/=-[i|]//=. rewrite !ren_inst. apply: pstep_subst. exact: A.
Qed.

Lemma pstep_compat m n (sigma tau : subst m n) s t :
  psstep sigma tau -> pstep s t -> pstep (inst sigma s) (inst tau t).
Proof with eauto using @pstep, @psstep_up.
  move=> A B. elim: B n sigma tau A => //={m s t}...
  move=> m s1 s2 t1 t2 u -> _ A _ B n sigma tau C.
  apply: (@pstep_beta _ _ (inst (upi tau) s2) _ (inst tau t2))...
  (*asimpl...*)
Admitted.

Lemma pstep_compat_beta n (s1 s2 : tm n.+1) t1 t2 :
  pstep s1 s2 -> pstep t1 t2 -> pstep (inst (t1 .: @var _) s1) (inst (t2 .: @var _) s2).
Proof.
  move=> A B. by apply: pstep_compat A => -[|].
Qed.

Ltac inv H := inversion H; clear H; subst.

Lemma rho_triangle n : triangle (@pstep n) rho.
Proof with eauto using @pstep.
  move=> s t. elim=> {n s t} //=...
  - move=> n s1 s2 t1 t2 u -> {u} _ A _ B. exact: pstep_compat_beta.
  - move=> n s1 s2 t1 t2 A ih1 _ ih2. case: s1 A ih1 => //=...
    move=> s A ih1. inv A; inv ih1...
Qed.

Theorem church_rosser n :
  CR (@step n).
Proof.
  apply: (cr_method (e2 := pstep) (rho := rho)).
    exact: step_pstep. exact: pstep_red. exact: rho_triangle.
Qed.
Hint Resolve church_rosser.

(** **** Reduction behaviour *)

Lemma normal_step_sort n u : normal (@step n) (univ n u).
Proof. move=> [s st]. inversion st. Qed.
Hint Resolve normal_step_sort.

CoInductive RedProdSpec {n} A1 B1 : tm n -> Prop :=
| RedProdSpecI A2 B2 : red A1 A2 -> red B1 B2 -> RedProdSpec A1 B1 (prod A2 B2).

Lemma red_prod_inv n (A1 : tm n) B1 C :
  red (prod A1 B1) C -> RedProdSpec A1 B1 C.
Proof.
  elim=> {C} [|C D _ [A2 B2]].
  - by constructor.
  - move=> ra12 rb12 st. inv st; constructor; eauto using star.
Qed.

Lemma inj_sort n u v : univ n u === univ n v -> u = v.
Proof.
  move=> A. suff: (univ n u = univ n v) => [[//]|].
  eapply cr_conv_normal => //; done.
Qed.

Lemma inj_prod n (A1 A2 : tm n) B1 B2 :
  prod A1 B1 === prod A2 B2 -> A1 === A2 /\ B1 === B2.
Proof.
  move=>/church_rosser[z /red_prod_inv s1 /red_prod_inv s2].
  inv s1; inv s2; split; eauto using join_conv.
Qed.

Lemma conv_prod_sort n A B u :
  ~(prod A B === univ n u).
Proof.
  move=> h. apply cr_star_normal in h => //. apply red_prod_inv in h. inv h.
Qed.

(** **** Cumulativity / Subtyping relation *)

Inductive sub1 {n} : tm n -> tm n -> Prop :=
| sub1_refl A : sub1 A A
| sub1_sort u v : u ≤ v -> sub1 (univ n u) (univ n v)
| sub1_prod A B1 B2 : @sub1 n.+1 B1 B2 -> sub1 (prod A B1) (prod A B2).

CoInductive sub {n} (A B : tm n) : Prop :=
| SubI A' B' : sub1 A' B' -> A === A' -> B' === B -> sub A B.
Infix "<:" := sub (at level 50, no associativity).

Lemma sub1_sub n (A B : tm n) : sub1 A B -> sub A B. move=> /SubI. exact. Qed.
Lemma sub1_conv n (B A C : tm n) : sub1 A B -> B === C -> A <: C. move=>/SubI. exact. Qed.
Lemma conv_sub1 n (B A C : tm n) : A === B -> sub1 B C -> A <: C. move=>c/SubI. exact. Qed.

Lemma conv_sub n (A B : tm n) : A === B -> A <: B.
Proof. move/conv_sub1. apply. exact: sub1_refl. Qed.

Lemma sub_refl n (A : tm n) : A <: A.
Proof. apply: sub1_sub. exact: sub1_refl. Qed.
Hint Resolve sub_refl.

Lemma sub_sort n u v : u ≤ v -> univ n u <: univ n v.
Proof. move=> leq. exact/sub1_sub/sub1_sort. Qed.

Lemma sub1_trans n (A B C D : tm n) :
  sub1 A B -> B === C -> sub1 C D -> A <: D.
Proof with eauto using @sub1, @sub1_sub, @sub1_conv, @conv_sub1.
  move=> sb. elim: sb C D => {n A B}
    [n A C D|n u v leq C D conv sb|n A B1 B2 sb1 ih C D conv sb2]...
  - inv sb...
    + apply: sub_sort. move: conv => /inj_sort eqn. subst.
      exact: le_trans leq _.
    + exfalso. exact: conv_prod_sort (conv_sym conv).
  - inv sb2...
    + exfalso. exact: conv_prod_sort conv.
    + move: conv => /inj_prod[conv1 conv2].
      move: (ih _ _ conv2 H) => {ih} sub. inv sub.
      eapply SubI. eapply sub1_prod... eapply conv_prod... exact: conv_prod.
Qed.

Lemma sub_trans n (B A C : tm n) :
  A <: B -> B <: C -> A <: C.
Proof.
  move=> [A' B' s1 c1 c2] [B'' C' s2 c3 c4]. move: (conv_trans _ c2 c3) => h.
  case: (sub1_trans s1 h s2) => A0 B0 s3 c5 c6. apply: (SubI s3).
  exact: conv_trans c5. exact: conv_trans c4.
Qed.

Lemma sub_prod_inv n (A1 A2 : tm n) B1 B2 :
  prod A1 B1 <: prod A2 B2 -> A1 === A2 /\ B1 <: B2.
Proof.
  move=> [A' B' []].
  - move=> C c1 c2. have{c1 c2}/inj_prod[c1 c2]: prod A1 B1 === prod A2 B2.
     exact: conv_trans c2.
    split=>//. exact: conv_sub.
  - by move=> u _ _ /conv_prod_sort[].
  - move=> A C1 C2 sb /inj_prod[c1 c2] /inj_prod[c3 c4]. split.
    exact: conv_trans c3. exact: SubI sb c2 c4.
Qed.

Lemma sub1_inst m n (sigma : subst m n) A B :
  sub1 A B -> sub1 (inst sigma A) (inst sigma B).
Proof. move=> s. elim: s n sigma => /=; eauto using @sub1. Qed.

Lemma sub_inst m n (sigma : subst m n) A B :
  A <: B -> inst sigma A <: inst sigma B.
Proof. move=> [A' B' /sub1_inst]; eauto using @sub, @conv_subst. Qed.

(** **** Typing *)

Definition ctx n := tm n -> tm n -> Prop.

(*Notation "Gamma `_ i" := (dget Gamma i).*)

Reserved Notation "[ Gamma |- ]".
Reserved Notation "[ Gamma |- s :- A ]".

Inductive ext {n} (Gamma : ctx n) (A : tm n) : ctx n.+1 :=
| ext_bound : ext Gamma A (var bound) (rename shift A)
| ext_shift s B : Gamma s B -> ext Gamma A (rename shift s) (rename shift B).

Inductive has_type {n} (Gamma : ctx n) : ctx n :=
| ty_var s A :
    Gamma s A ->
    [ Gamma |- s :- A ]
| ty_sort u :
    [ Gamma |- univ n u :- univ n (sort_succ u) ]
| ty_app A B s t :
    [ Gamma |- s :- prod A B ] ->
    [ Gamma |- t :- A ] ->
    [ Gamma |- app s t :- inst (t .: @var _) B ]
| ty_lam A B s u :
    [ Gamma |- A :- univ n u ] ->
    [ ext Gamma A |- s :- B ] ->
    [ Gamma |- lam s :- prod A B ]
| ty_prod A B u v :
    [ Gamma |- A :- univ n u ] ->
    [ ext Gamma A |- B :- univ n.+1 v ] ->
    [ Gamma |- prod A B :- univ n (sort_prod u v) ]
| ty_sub u A B s :
    A <: B ->
    [ Gamma |- B :- univ n u ] ->
    [ Gamma |- s :- A ] ->
    [ Gamma |- s :- B ]
where "[ Gamma |- s :- A ]" := (@has_type _ Gamma s A).

Definition is_type {n} (Gamma : ctx n) (A : tm n) :=
  exists u, Gamma A (univ n u).

Definition context_ok {n} (Gamma : ctx n) :=
  forall s A, Gamma s A -> is_type (has_type Gamma) A.

(*Lemma ty_evar Gamma (A : term) (x : var) :
  A = Gamma`_x -> x < size Gamma -> [ Gamma |- Var x :- A ].
Proof. move->. exact: ty_var. Qed.*)

Lemma ty_eapp n (Gamma : ctx n) A B C s t :
  C = inst (t .: @var _) B ->
  [ Gamma |- s :- prod A B ] -> [ Gamma |- t :- A ] ->
  [ Gamma |- app s t :- C ].
Proof. move=>->. exact: ty_app. Qed.

(* Type well-formedness *)

Notation "[ Gamma |- s ]" := (is_type (has_type Gamma) s). (*(exists n, [ Gamma |- s :- Sort n ]).*)

Lemma ty_sort_wf n Gamma u : [ Gamma |- univ n u ].
Proof. exists (sort_succ u). exact: ty_sort. Qed.
Hint Resolve ty_sort_wf ty_sort.

Lemma ty_prod_wf n (Gamma : ctx n) A B :
  [ Gamma |- A ] -> [ ext Gamma A |- B ] -> [ Gamma |- prod A B ].
Proof.
  move=> [u tp1] [v tp2]. exists (sort_prod u v). apply: ty_prod.
  - eapply (ty_sub (A := univ _ u)); eauto=> //.
  - eapply (ty_sub (A := univ _ v)); eauto.
Qed.

(* Substitution Lemma *)

Definition cmorph {m n} (f : subst m n) (Gamma : ctx m) (Delta : ctx n) :=
  forall s A, Gamma s A -> Delta (inst f s) (inst f A).

Notation "[ Gamma |- sigma -| Delta ]" :=
  (cmorph sigma Gamma Delta).
  (*forall s A, Delta s A -> Gamma (inst sigma s) (inst sigma A)).*)
  (*(forall x, x < size Gamma -> [ Delta |- sigma x :- (Gamma`_x).[sigma] ]).*)

Lemma tysubst_ids m (Gamma : ctx m) :
  [ Gamma |- @var _ -| Gamma ].
Proof.
  move=> s A e. rewrite !inst_ids. exact: e.
Qed.

Lemma tysubst_var m (Gamma : ctx m) :
  [ Gamma |- @var _ -| has_type Gamma ].
Proof.
  move=> s A e. rewrite !inst_ids. exact: ty_var.
Qed.

Lemma tysubst_comp m n k (Gamma : ctx m) (Delta : ctx n) (Theta : ctx k) (f : subst m n) (g : subst n k) :
  [ Gamma |- f -| Delta ] -> [ Delta |- g -| Theta ] -> [ Gamma |- f >> inst g -| Theta ].
Proof.
  move=> h1 h2 s A /h1/h2. by rewrite !inst_comp.
Qed.

Lemma tysubst_ren_comp m n k (Gamma : ctx m) (Delta : ctx n) (Theta : ctx k) (f : subst m n) (g : ren n k) :
  [ Gamma |- f -| Delta ] -> [ Delta |- g >> @var _ -| Theta ] -> [ Gamma |- f >> rename g -| Theta ].
Proof.
  move=> /tysubst_comp h/h{h}h s A e. rewrite -!inst_ren_comp !ren_inst !inst_comp. exact: h.
Qed.

Lemma tysubst_shift m (A : tm m) (Gamma : ctx m) :
  [ Gamma |- shift >> @var _ -| ext Gamma A ].
Proof.
  move=> s B e. rewrite -!ren_inst. apply: ext_shift. exact: e.
Qed.

Lemma tysubst_cons m n (sigma : subst m n) (Gamma : ctx m) (Delta : ctx n) (s : tm n) (A : tm m) :
  Delta s (inst sigma A) -> [ Gamma |- sigma -| Delta ] -> [ ext Gamma A |- s .: sigma -| Delta ].
Proof.
  move=> h1 h2 t B e. inv e=> /=; rewrite !ren_inst_comp; fsimpl.
  exact: h1. apply: h2. exact: H.
Qed.

Lemma tysubst_up m n (sigma : subst m n) (Gamma : ctx m) (Delta : ctx n) A :
  [ Gamma |- sigma -| Delta ] -> [ ext Gamma A |- upi sigma -| ext Delta (inst sigma A) ].
Proof.
  move=> h. unfold upi. apply: tysubst_cons. rewrite -inst_ren_comp. apply: ext_bound.
  apply: tysubst_ren_comp h _. exact: tysubst_shift.
Qed.

Lemma ty_map m n (sigma : subst m n) (Gamma : ctx m) (Delta : ctx n) s A :
  [ Gamma |- s :- A ] ->
  [ Gamma |- sigma -| Delta ] ->
  [ Delta |- inst sigma s :- inst sigma A ].
Proof.
  move=> tp. elim: tp n sigma Delta => {m Gamma s A} /=
    [m Gamma t B si|m Gamma u|m Gamma A B s t _ ih1 _ ih2|
     m Gamma A B s u _ ih1 _ ih2|m Gamma A B u v _ ih1 _ ih2|
     m Gamma u A B s sub _ ih1 _ ih2]
    n sigma Delta tp //=.
  - apply: ty_var. exact: tp.
  - apply: ty_eapp (ih1 _ _ _ tp) (ih2 _ _ _ tp). admit.
  - apply: ty_lam. exact: ih1. apply: ih2. exact: tysubst_up.
  - apply: ty_prod. exact: ih1. apply: ih2. exact: tysubst_up.
  - apply: ty_sub; last first. exact: ih2. exact: ih1. exact: sub_inst.
Admitted.

(*Lemma ty_renaming xi Gamma Delta s A :
  [ Gamma |- s :- A ] ->
  (forall x, x < size Gamma -> xi x < size Delta) ->
  (forall x, x < size Gamma -> (Gamma`_x).[ren xi] = Delta`_(xi x)) ->
  [ Delta |- s.[ren xi] :- A.[ren xi] ].
Proof.
  move=> tp. elim: tp xi Delta => {Gamma s A} /=
    [Gamma x si|Gamma n|Gamma A B s t _ ih1 _ ih2|
     Gamma A B s n _ ih1 _ ih2|Gamma A B n _ ih1 _ ih2|
     Gamma n A B s sub _ ih1 _ ih2]
    xi Delta subctx eqn //=.
  - rewrite eqn //. apply: ty_var. exact: subctx.
  - apply: (@ty_eapp _ A.[ren xi] B.[up (ren xi)]). autosubst.
      exact: ih1. exact: ih2.
  - eapply ty_lam. by eapply ih1. asimpl. apply: ih2.
    + by move=> [//|x /subctx].
    + by move=> [_|x /eqn/=<-]; autosubst.
  - apply: ty_prod. exact: ih1. asimpl. apply: ih2.
    + by move=> [//|x /subctx].
    + by move=> [_|x /eqn/=<-]; autosubst.
  - apply: (@ty_sub _ n A.[ren xi] B.[ren xi]). exact: sub_subst.
      exact: ih1. exact: ih2.
Qed.*)

Lemma weakening n (Gamma : ctx n) s A B :
  [ Gamma |- s :- A ] -> [ ext Gamma B |- rename shift s :- rename shift A ].
Proof. rewrite !ren_inst => /ty_map. apply. exact: tysubst_shift. Qed.

Lemma eweakening n (Gamma : ctx n) s s' A A' B :
  s' = rename shift s -> A' = rename shift A ->
  [ Gamma |- s :- A ] -> [ ext Gamma B |- s' :- A' ].
Proof. move=>->->. exact: weakening. Qed.

(*Lemma ty_ok Gamma :
  [ Gamma |- ] -> forall x, x < size Gamma -> [ Gamma |- Gamma`_x ].
Proof.
  elim=> // {Gamma} Gamma A n tp _ ih [_|x /ih [{n tp} n tp]];
    exists n; exact: weakening tp.
Qed.*)

Lemma tysubst_upi m n (sigma : subst m n) (Gamma : ctx m) (Delta : ctx n) A :
  [ Gamma |- sigma -| has_type Delta ] ->
  [ ext Gamma A |- upi sigma -| has_type (ext Delta (inst sigma A)) ].
Proof.
  move=> h. unfold upi. apply: tysubst_cons. rewrite -inst_ren_comp. apply: ty_var. exact: ext_bound.
  apply: tysubst_ren_comp h _. move=> s B /weakening e. rewrite -!ren_inst. exact: e.
Qed.

(*Lemma sty_up sigma Gamma Delta A :
  [ Delta |- sigma -| Gamma ] ->
  [ A.[sigma] :: Delta |- up sigma -| A :: Gamma ].
Proof.
  move=> stp [_//|x /stp tp] /=. apply: ty_evar => //=. autosubst.
  apply: eweakening tp; autosubst.
Qed.*)

Lemma ty_subst m n (sigma : subst m n) Gamma Delta s A :
  [ Gamma |- sigma -| has_type Delta ] -> [ Gamma |- s :- A ] ->
  [ Delta |- inst sigma s :- inst sigma A ].
Proof.
  move=> stp tp. elim: tp n sigma Delta stp => {m Gamma s A} /=
    [m Gamma t B si|m Gamma u|m Gamma A B s t _ ih1 _ ih2|
     m Gamma A B s u _ ih1 _ ih2|m Gamma A B u v _ ih1 _ ih2|
     m Gamma u A B s sub _ ih1 _ ih2]
    n sigma Delta tp //=.
  - exact: tp.
  - apply: ty_eapp; last first. exact: ih2. exact: ih1. admit.
  - eapply ty_lam. by eapply ih1. apply: ih2. exact: tysubst_upi.
  - apply: ty_prod. exact: ih1. apply ih2. exact: tysubst_upi.
  - apply: ty_sub; last first. exact: ih2. exact: ih1. exact: sub_inst.
Admitted.

Lemma ty_cut n (Gamma : ctx n) s t A B :
  [ ext Gamma A |- s :- B ] -> [ Gamma |- t :- A ] ->
  [ Gamma |- inst (t .: @var _) s :- inst (t .: @var _) B ].
Proof.
  move=> /ty_subst h tp. apply: h. apply: tysubst_cons; last first.
  exact: tysubst_var. by rewrite inst_ids.
Qed.

Lemma ty_ctx_conv n (Gamma : ctx n) s A B C u :
  [ ext Gamma A |- s :- C ] -> B === A -> [ Gamma |- A :- univ n u ] ->
  [ ext Gamma B |- s :- C ].
Proof.
  move=> tp1 conv tp2. cut ([ ext Gamma B |- inst (@var _) s :- inst (@var _) C ]). by rewrite !inst_ids.
  apply: ty_subst tp1. move=> t T [|]{t T}.
  - rewrite ren_inst_comp /=. rewrite -ren_inst. apply: (@ty_sub _ _ _ (rename shift B)).
    + rewrite !ren_inst. apply: sub_inst. exact: conv_sub.
    + exact: eweakening tp2.
    + apply: ty_var. exact: ext_bound.
  - move=> t T e. rewrite !inst_ids. apply: ty_var. exact: ext_shift.
Qed.

(* Finite contexts *)

Definition cempty {n} : ctx n :=
  fun _ _ => False.

Inductive is_finite : forall {n}, ctx n -> Prop :=
| is_finite_empty n : @is_finite n cempty
| is_finite_ext n (A : tm n) (Gamma : ctx n) :
    is_finite Gamma -> is_finite (ext Gamma A).

Lemma is_finite_var {n} (Gamma : ctx n) :
  is_finite Gamma ->
  forall s A, Gamma s A -> exists i, s = var i.
Proof.
  elim=> {n Gamma}//n A Gamma _ ih s B [{s B}|{s B}s B tp]. by exists bound.
  case: (ih s B tp) => i -> /=. by exists (shift i).
Qed.

Lemma is_finite_prod {n} (Gamma : ctx n) A B C :
  is_finite Gamma -> ~Gamma (prod A B) C.
Proof.
  by move=> /is_finite_var H/H[].
Qed.

Lemma is_finite_lam {n} (Gamma : ctx n) s A :
  is_finite Gamma -> ~Gamma (lam s) A.
Proof.
  by move=> /is_finite_var H/H[].
Qed.

(* Inversion Lemmas *)

Lemma ty_prod_invX n (Gamma : ctx n) A B u :
  is_finite Gamma ->
  [ Gamma |- prod A B :- u ] ->
    [ Gamma |- A ] /\ [ ext Gamma A |- B ].
Proof.
  move e:(prod A B) => s tc tp. elim: tp A B e tc => //{n Gamma s u}.
  - move=> n Gamma s A tp A' B' eqn tc; subst. exfalso. exact: is_finite_prod tc tp.
  - move=> n Gamma A B u v tp1 _ tp2 _ A' B' [e1 e2] tc; subst. split; eexists; by eauto.
Qed.

Lemma ty_prod_inv n (Gamma : ctx n) A B :
  is_finite Gamma ->
  [ Gamma |- prod A B ] ->
    [ Gamma |- A ] /\ [ ext Gamma A |- B ].
Proof. move=> tc [u h]. exact: ty_prod_invX h. Qed.

Lemma ty_lam_invX n (Gamma : ctx n) s A B C :
  is_finite Gamma ->
  [ Gamma |- lam s :- C ] ->
  (C <: prod A B /\ [ ext Gamma A |- B ]) \/ prod A B = C ->
  [ ext Gamma A |- s :- B ].
Proof.
  move e:(lam s) => t tc tp. elim: tp A B s e tc => // {n Gamma t C}.
  - move=> n Gamma s A tp A' B' t eqn tc. subst. exfalso. exact: is_finite_lam tc tp.
  - move=> n Gamma A B s u tp1 _ tp2 _ A' B' s' [->] tc [|[->->//]].
    move=> [/sub_prod_inv[con sub] [m tp3]]. apply: ty_sub sub tp3 _.
    apply: ty_ctx_conv tp2 _ tp1. exact: conv_sym.
  - move=> n Gamma u A B s sub1 tp1 ih1 tp2 ih2 A' B' t eqn1 tc [[sub2 [m tp3]]|eqn2];
      subst; apply: ih2 => //; left; split => //.
    + exact: sub_trans sub1 sub2. by exists m.
    + move: tc => /ty_prod_inv H. case: (H A' B') => //. by exists u.
Qed.

Lemma ty_lam_inv n (Gamma : ctx n) s A B :
  is_finite Gamma ->
  [ Gamma |- lam s :- prod A B ] -> [ ext Gamma A |- s :- B ].
Proof. move=> tc tp. apply: ty_lam_invX tp _ => //. by right. Qed.

(* Main Result *)

Lemma is_type_weakening n (Gamma : ctx n) A B :
  [ Gamma |- B ] -> [ ext Gamma A |- rename shift B ].
Proof.
  move=> [u tp]. exists u. apply: eweakening => //; last first. exact: tp. by [].
Qed.

Lemma context_ok_ext n (Gamma : ctx n) A :
  context_ok Gamma -> [Gamma |- A] -> context_ok (ext Gamma A).
Proof.
  move=> h1 h2 s B [{s B}|{s B}s B /h1 h3]; exact: is_type_weakening.
Qed.

Lemma propagation n (Gamma : ctx n) s A :
  is_finite Gamma -> context_ok Gamma ->
  [ Gamma |- s :- A ] -> [ Gamma |- A ].
Proof.
  move=> tc1 tc2 tp. elim: tp tc1 tc2 => {n Gamma s A} /=
    [n Gamma s A e tc1 tc2||n Gamma A B _ t _ ih tp _ tc1 tc2|n Gamma A B s u v h1 h2 h3 h4 tc|
    |n Gamma u A B s _ tp _ _ _ _ tc] //.
  - exact: tc2 e.
  - move: (ih tc1 tc2) => [u {ih} H]. case: (ty_prod_inv tc1 (ex_intro _ _ H)) => _ [v ih].
    exists v. exact: (@ty_cut _ _ _ _ _ _ ih tp).
  - apply: ty_prod_wf. by exists u. apply: h3. exact: is_finite_ext. apply: context_ok_ext => //. by exists u.
  - by exists u.
Qed.

Lemma subject_reduction n (Gamma : ctx n) s A :
  [ Gamma |- s :- A ] ->
  is_finite Gamma -> context_ok Gamma ->
    forall t, step s t -> [ Gamma |- t :- A ].
Proof.
  elim => {n Gamma s A}.
  - move=> n Gamma s A e tc1 tc2 t st. case: (is_finite_var tc1 e) => i E; subst. by inv st.
  - move=> n Gamma u tc1 tc2 t st. by inv st.
  - move=> n Gamma A B s t tp1 ih1 tp2 ih2 tc1 tc2 u st. inv st.
    + apply ty_lam_inv in tp1 => //. exact: ty_cut tp1 tp2.
    + apply: ty_app tp2. exact: ih1.
    + move: (tp1) => /propagation. move/(_ tc1 tc2) => /ty_prod_inv[]// _ [u tp3].
      eapply ty_sub; last first. eapply ty_app. eassumption. exact: ih2.
      move: (ty_cut tp3 tp2) => /= h. apply h. apply: conv_sub.
      apply: conv_beta. exact: conv1i.
  - move=> n Gamma A B s u tp1 ih1 tp2 ih2 tc1 tc2 t st. inv st.
    apply: ty_lam (tp1) _. apply: ih2 => //. exact: is_finite_ext. apply: context_ok_ext => //. by exists u.
  - move=> n Gamma A B u v tp1 ih1 tp2 ih2 tc1 tc2 t st. inv st.
    + apply: ty_prod. exact: ih1. eapply (ty_ctx_conv tp2).
      exact: conv1i. eassumption.
    + apply: ty_prod => //. apply: ih2 => //. exact: is_finite_ext. apply: context_ok_ext => //. by exists u.
  - move=> n Gamma s A B u tp1 ih1 con tp2 ih2 tc1 tc2 t st.
    apply: ty_sub tp1 ih1 _. exact: ih2.
Qed.
