Require Import base data.fintype.
(*Require Import cat.category cat.functor cat.nat cat.products cat.products cat.exponentials cat.yoneda cat.presheaf.*)
(*Require Import data.fintype.*)
Require Import pts.category.
Set Universe Polymorphism.
Set Printing Universes.

(** ** Presheaf models of lambda terms *)

Polymorphic Cumulative Structure model@{i j k} := mk_model {
  context_cat : cartcat@{i j};
  domain :> presheaf@{i j k} context_cat;
  alpha  : pnat (PProd@{i j k i j k} domain domain) domain;
  lambda : pnat (PExp@{i j k} domain domain) domain
}.

(** ** Syntactic model of lambda terms *)
Notation ℕ := Datatypes.nat.

Inductive tm (n : ℕ) : Type :=
| var (x : fin n) : tm n
| app (s t : tm n) : tm n
| lam (s : tm n.+1) : tm n.
Arguments var {n} x.

Fixpoint rename {m n : ℕ} (f : ren m n) (s : tm m) : tm n :=
  match s with
  | var x => var (f x)
  | app s t => app (rename f s) (rename f t)
  | lam s => lam (rename (up f) s)
  end.

Definition subst m n := fin m -> tm n.

Definition uptm {m n} (f : subst m n) : subst m.+1 n.+1 :=
  var bound .: f >> rename shift.

Fixpoint inst {m n} (f : subst m n) (s : tm m) : tm n :=
  match s with
  | var x => f x
  | app s t => app (inst f s) (inst f t)
  | lam s => lam (inst (uptm f) s)
  end.

(** *** Renaming lemmas *)

Lemma rename_id m (s : tm m) : rename id s = s.
Proof.
  elim: s => {m}[m x//|m/=s->t->//|m s ih/=].
  rewrite/up. fsimpl. by rewrite ih.
Qed.

Lemma rename_inst m n (f : ren m n) (s : tm m) :
  rename f s = inst (f >> var) s.
Proof.
  elim: s n f => {m}[m x//|m/=s ihs t iht|m s ih/=] n f.
    by rewrite ihs iht. rewrite/up/uptm ih. by fsimpl.
Qed.

Lemma rename_inst_comp m n k (f : ren m n) (g : subst n k) (s : tm m) :
  inst g (rename f s) = inst (f >> g) s.
Proof.
  elim: s n k f g => {m}/=[m x//|m s ihs t iht|m s ih] n k f g.
    by rewrite ihs iht. rewrite ih /up/uptm. by fsimpl.
Qed.

Lemma rename_comp m n k (f : ren m n) (g : ren n k) (s : tm m) :
  rename g (rename f s) = rename (f >> g) s.
Proof.
  by rewrite rename_inst rename_inst_comp rename_inst.
Qed.

(** *** Instantiation lemmas *)

Lemma inst_id m (s : tm m) :
  inst var s = s.
Proof.
  by rewrite -{2}[s]rename_id rename_inst.
Qed.

Lemma inst_rename_comp m n k (f : subst m n) (g : ren n k) (s : tm m) :
  rename g (inst f s) = inst (f >> rename g) s.
Proof.
  elim: s n k f g => {m}/=[m x//|m s ihs t iht|m s ih] n k f g.
    by rewrite ihs iht. rewrite ih /up/uptm. fsimpl=>/=. f_equal.
    f_equal. f_equal. f_equal. fext=> t /=. rewrite !rename_comp. by fsimpl.
Qed.

Lemma uptm_comp m n k (f : subst m n) (g : subst n k) :
  uptm f >> inst (uptm g) = uptm (f >> inst g).
Proof.
  fext=>/=-[i|]//=. by rewrite rename_inst_comp inst_rename_comp.
Qed.

Lemma inst_comp m n k (f : subst m n) (g : subst n k) (s : tm m) :
  inst g (inst f s) = inst (f >> inst g) s.
Proof.
  elim: s n k f g => {m}/=[m x//|m s ihs t iht|m s ih] n k f g.
    by rewrite ihs iht. by rewrite ih uptm_comp.
Qed.

(** *** The category of (term) contexts *)

Definition ctx : Type := ℕ.
Definition cmorph (m n : ctx) : Type := subst n m.
Definition cid (m : ctx) : cmorph m m := var.
Definition ccomp m n k (f : cmorph m n) (g : cmorph n k) : cmorph m k := g >> inst f.

Lemma ccomp_idL m n (f : cmorph m n) : ccomp (@cid _) f = f.
Proof. fext=>i. exact: inst_id. Qed.

Lemma ccomp_idR m n (f : cmorph m n) : ccomp f (@cid _) = f.
Proof. by []. Qed.

Lemma ccomp_assoc m n k l (f : cmorph m n) (g : cmorph n k) (h : cmorph k l) :
  ccomp (ccomp f g) h = ccomp f (ccomp g h).
Proof. fext=>i. symmetry. exact: inst_comp. Qed.

Canonical ctx_catMixin := @CatMixin ctx cmorph cid ccomp
  ccomp_idL ccomp_idR ccomp_assoc.
Canonical CTX := Eval hnf in Cat ctx cmorph ctx_catMixin.

(** *** Products in contexts *)

Definition ctx_prod (m n : ctx) : ctx := m + n.
Definition ctx_terminal : ctx := 0.

Fixpoint fin_split m : forall n, fin (m + n) -> fin m + fin n :=
  match m as m return forall n, fin (m+n) -> (fin m + fin n) with
  | 0 => fun n i => inr i
  | m.+1 => fun n i =>
    match i with
    | None => inl bound
    | Some j =>
      match @fin_split m n j with
      | inl k => inl (shift k)
      | inr k => inr k
      end
    end
  end.
(*Definition fin_split (m n : ℕ) (i : fin (n + m)) : fin n + fin m.
  induction n; cbn in *. right. exact i. destruct i as [j|].
  destruct (IHn j) as [L|R]. left. apply shift. exact L. right. exact R.
  left. exact bound.
Defined.*)

Definition fin_inl m n (i : fin m) : fin (m + n).
Proof.
  induction m; cbn in *. destruct i. destruct i as [j|].
  apply shift. exact (IHm j). exact bound.
Defined.

Definition fin_inr m n (i : fin n) : fin (m + n).
Proof.
  induction m; cbn in *. exact i. apply shift. exact IHm.
Defined.

Lemma fin_inl_split m n (i : fin m) :
  fin_split (@fin_inl m n i) = inl i.
Proof.
  induction m; cbn in *. destruct i. destruct i as [j|]; cbn=>//.
    by rewrite (IHm j).
Qed.

Lemma fin_inr_split m n (i : fin n) :
  fin_split (@fin_inr m n i) = inr i.
Proof.
  induction m; cbn in * => //. by rewrite IHm.
Qed.

Lemma fin_split_eta m n (i : fin (m + n)) :
  match fin_split i with
  | inl j => fin_inl n j
  | inr j => fin_inr m j
  end = i.
Proof.
  induction m; cbn in *=>//. destruct i as [j|]; cbn=>//.
  move: (IHm j). case: (fin_split j) => k eqn.
  cbn. by rewrite -eqn. by rewrite -eqn.
Qed.

Definition ctx_pi1 (m n : ctx) : hom (ctx_prod m n) m.
  move=> i. cbn. apply: var. exact: fin_inl.
Defined.

Definition ctx_pi2 (m n : ctx) : hom (ctx_prod m n) n.
  move=> i. cbn. apply: var. exact: fin_inr.
Defined.

Definition ctx_pair (m n k : ctx) (f : hom m n) (g : hom m k) : hom m (ctx_prod n k).
  move=> i. cbn. destruct (fin_split i) as [j|j]. exact (f j). exact (g j).
Defined.

Definition ctx_pair_pi1 (m n k : ctx) (f : hom m n) (g : hom m k) :
  ctx_pair f g ∘ @ctx_pi1 _ _ = f.
Proof.
  fext=> i /=. cbn. unfold ctx_pair. by rewrite fin_inl_split.
Qed.

Definition ctx_pair_pi2 (m n k : ctx) (f : hom m n) (g : hom m k) :
  ctx_pair f g ∘ @ctx_pi2 _ _ = g.
Proof.
  fext=> i /=. cbn. unfold ctx_pair. by rewrite fin_inr_split.
Qed.

Definition ctx_pair_eta (m n k : ctx) (f : hom m (ctx_prod n k)) :
  ctx_pair (f ∘ @ctx_pi1 _ _) (f ∘ @ctx_pi2 _ _) = f.
Proof.
  fext=> i /=. cbn. unfold ctx_pair. rewrite -{2}[i]fin_split_eta.
    by case: (fin_split i).
Qed.

Definition ctx_bang (n : ctx) : hom n ctx_terminal := null.

Lemma ctx_terminal_eta (n : ctx) (f : hom n ctx_terminal) :
  f = ctx_bang n.
Proof. by fext. Qed.

Canonical ctx_cartcatMixin :=
  @CartCat.Mixin ctx cmorph ctx_catMixin
    ctx_terminal ctx_bang ctx_prod ctx_pi1 ctx_pi2 ctx_pair
    ctx_pair_pi1 ctx_pair_pi2 ctx_pair_eta
    ctx_terminal_eta.
Canonical ctx_cartcat :=
  Eval hnf in CartCat.pack ctx_cartcatMixin.

(** *** Terms form a presheaf on the category of contexts *)

(* Definition TM = y 1 *)
(*Definition TM : presheaf CTX := y (1%nat : CTX).*)
Program Definition Tm@{i j k} : presheaf@{i j k} ctx_cartcat :=
  mk_presheaf (fun n => tm n) (fun m n f s => inst f s) _ _.
Next Obligation.
  exact: inst_id.
Qed.
Next Obligation.
  exact: inst_comp.
Qed.

(** *** Term model *)

Program Definition APP : pnat (PProd Tm Tm) Tm :=
  mk_pnat (fun n p => app p.1 p.2) (fun _ _ _ _ => erefl).

Program Definition LAM : pnat (PExp Tm Tm) Tm :=
  mk_pnat (fun n F => lam (F n.+1 (shift >> var) (var bound))) _.
Next Obligation.
  move: (svalP x) => /=->. cbn. repeat f_equal. fext=> i.
  cbn. by rewrite rename_inst.
Qed.

Definition TM@{i j k} : model := @mk_model@{i j k} ctx_cartcat Tm APP LAM.

(** ** Morphism from lambda terms *)

Section Evaluation.
  Variable D : model.

  Fixpoint times {A : cartcat} (D : A) (n : nat) : A :=
    match n with
    | 0 => 1
    | m.+1 => times D m × D
    end.

  Definition times_pi {A : cartcat} (C : A) (n : nat) : fin n -> hom (times C n) C.
    induction n. by []. cbn. destruct 1. refine (π₁ ∘ _). exact: IHn. exact π₂.
  Defined.

  Definition eval {n} (s : tm n) : pnat (times (domain D) n) (domain D).
    induction s.
    exact (times_pi (domain D) x). (*apply (proj x).*)
    refine (_ ∘ alpha D). refine (⟨_,_⟩). exact IHs1. exact IHs2.
    refine (_ ∘ lambda D). apply nat_lam. cbn in IHs. exact IHs.
  Defined.
