Require Import base ord tower.tarski tower.associated_closure tower.tower.

Lemma prod_exEc T (X : clat) I (P : I -> Prop) (F : I -> T -> X) (x : T) :
  (∃ i | P i, F i) x = (∃ i | P i, F i x).
Proof.
  rewrite prod_exE. f_equal; fext=> i. exact: prod_exE.
Qed.

(** ** Labeled transition systems and similarity *)

Structure lts (Σ : Type) := mk_lts
  { lts_nodes :> Type
  ; step : Σ -> Rel lts_nodes
  }.
Arguments mk_lts {Σ} lts_nodes step.

Section Similarity.
Variable Σ : Type.
Variable (A B : lts Σ).

Definition sim (R : A -> B -> Prop) (a : A) (b : B) :=
  forall σ a', step σ a a' -> exists2 b', step σ b b' & R a' b'.
Definition similar := gfp sim.

Lemma similar_def (a : A) (b : B) :
  similar a b = exists2 R : A -> B -> Prop, (forall a' b', R a' b' -> sim R a' b') & R a b.
Proof.
  by rewrite/similar gfp_def !prod_exEc prop_exEc.
Qed.

End Similarity.

(** ** Failure of cocontinuouity *)

Definition loop : lts unit :=
  mk_lts unit (fun _ _ _ => True).

Definition countable_branching : lts unit :=
  mk_lts (option nat) (fun _ from to =>
    match from, to with
    | None, Some _ => True
    | Some n, Some m => n = m.+1
    | _, _ => False
    end).

Definition nsim {Σ} {A B : lts Σ} (n : nat) : A -> B -> Prop :=
  iter n (@sim Σ A B) (fun _ _ => True).

Lemma loop_countable_branching_nsim n :
  @nsim unit loop countable_branching n tt None.
Proof.
  case: n => //= n [] [] _. exists (Some n). exact I.
  elim: n => //= n ih [] [] _. exists (Some n). by []. exact ih.
Qed.

Lemma loop_countable_branching_not_similar :
  ~@similar unit loop countable_branching tt None.
Proof.
  rewrite similar_def.
  case=> R rs /rs/(_ tt tt I)[/=[]//n _]. elim: n.
  - by case/rs/(_ tt tt I) => /=-[].
  - move=> n ih /rs/(_ tt tt I) -[/=[]//m[<-{m}]]. exact ih.
Qed.

(** ** Similarity at Type *)

(*Section TypeSimilarity.
  Variables (Σ : Type) (A B : lts Σ).
  Definition tsim (R : A -> B -> Type) a b : Type :=
    forall σ a', step σ a a' -> { b' : B & (step σ b b' * R a' b')%type }.

  CoInductive dependent_back_sim (R : A -> B -> Type) (F : forall σ a a' b, R a b -> step σ a a' -> B) (a' : A) : B -> Type :=
  | DepBackSim σ a b (r : R a b) (st : step σ a a') :
      @dependent_back_sim R F a' (F σ a a' b r st).

  CoInductive tsimilarity (R : A -> B -> Type) (a : A) (b : B) : Type :=
  | TSimilar :
      (forall (η : forall σ c c' d, R c d -> step σ c c' -> { d' & step σ d d' }) σ a',
          step σ a a' ->
          { b' &
            (step σ b b' *
             tsimilarity
               (@dependent_back_sim R (fun σ c c' d r st => (η σ c c' d r st).1)) a' b')%type }) ->
      tsimilarity R a b.

  Lemma tsimilarity_coind F :
    (forall R a b, F R a b ->
       forall (η : forall σ c c' d, R c d -> step σ c c' -> { d' & step σ d d' }) σ a',
         step σ a a' ->
         { b' &
           (step σ b b' *
            F (@dependent_back_sim R (fun σ c c' d r st => (η σ c c' d r st).1)) a' b')%type }) ->
    forall R a b, F R a b -> tsimilarity R a b.
  Proof.
    move=> H. cofix co. move=> R a b fab. constructor=> η σ a' st.
    destruct (H R a b fab η σ a' st) as [b' [st' H']]. exists b'.
    split. exact st'. apply co. exact H'.
  Defined.

  Lemma tsimilarity_map (R S : A -> B -> Type) :
    (forall a b, R a b -> S a b) ->
    (forall a b, tsimilarity R a b -> tsimilarity S a b).
  Proof.
    move: R S. cofix tmap => R S le_R_S a b [H]. constructor.
    move=> η σ a' sta. case: (H (fun τ c c' d rcd st => η τ c c' d (le_R_S c d rcd) st) σ a' sta).
    move=> b' [stb Hb]. exists b'. split. exact stb. move: a' b' {σ sta stb} Hb.
    apply: tmap. move=> a' b' /=[]{b'}σ aa b' r st /=.
    set P := fun _ => _. exact: (@DepBackSim S P a' σ aa b').
  Defined.

  Lemma tsimilarity_compat (R : A -> B -> Type) a b :
    tsimilarity (tsim R) a b -> tsim (tsimilarity R) a b.
  Proof.
    move=> [tab] σ a' sta.
    case: (tab (fun σ c c' d tcd st => let (d', st) := tcd σ c' st in (d'; st.1)) σ a' sta).
    move=> b' [stb tab']. exists b'. split. exact stb. move: a' b' {a b tab σ sta stb} tab'.
    apply: tsimilarity_map => a' b' [] {b'} σ a b r sta. case: (r σ a') => b' [stb H] /=. exact H.
  Defined.

  Variable F : (A -> B -> Type) -> A -> B -> Type.
  Hypothesis F_map :
    forall (R S : A -> B -> Type),
      (forall a b, R a b -> S a b) ->
      (forall a b, F R a b -> F S a b).
  Hypothesis F_compat :
    forall (R : A -> B -> Type) a b, F (tsim R) a b -> tsim (F R) a b.

  Lemma tsimilarity_companion (R : A -> B -> Type) a b :
    F R a b -> tsimilarity R a b.
  Proof.
    move: R a b. apply: tsimilarity_coind => R a b fab η σ a' sta.
    set G := dependent_back_sim _. apply: (@F_compat G a b _ σ a' sta).
    move: a b {σ sta a'} fab. apply: F_map => a b rab σ a' sta.
    exists (η σ a a' b rab sta).1. split. by case: (η σ a a' b rab sta).
    exact: DepBackSim.
  Defined.
End TypeSimilarity.
*)
